/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 0.1
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: P. Dechamboux, Y. Devaux, F. Dhumeaux.
 *
 */

package org.objectweb.speedo.generation.enhancer;

import org.objectweb.speedo.api.SpeedoEnhancerException;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.metadata.SpeedoPackage;
import org.objectweb.speedo.metadata.SpeedoXMLDescriptor;
import org.objectweb.speedo.metadata.SpeedoIdentity;
import org.objectweb.speedo.generation.lib.NamingRules;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassWriter;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collection;
import java.util.List;
import java.util.TreeSet;
import java.util.Set;

/**
 * Enhances a set of Java classes and renames it with <code>Delegate</code> as
 * suffix. <p>The modifications to be done are described by the JDO
 * Implementation's rules.</p>
 *
 * @author	Y. Devaux
 */
public class AccessEnhancer extends EnhancerComponent {

    public final static String LOGGER_NAME
            = "org.objectweb.speedo.generation.enhancer.access";

    /**
     * Initializes this AccessEnhancer
     */
    public void init() {
        logger = scp.loggerFactory.getLogger(LOGGER_NAME);
        //TODO: manage classes localized in a jar file
        isSrcJar = false;
    }

    /**
     * Loads all binary classes described by the Object Model and applies
     * revelant modification to each of them.
     *
     * @exception SpeedoEnhancerException if something goes wrong
     */
    public void process() throws SpeedoException {

        if (scp.xmldescriptor == null || scp.xmldescriptor.isEmpty())
            throw new SpeedoEnhancerException("xmldescriptor not initialised");

        ArrayList except = new ArrayList();
        Collection xmls = scp.xmldescriptor.values();

        for (Iterator itDesc = xmls.iterator(); itDesc.hasNext();) {
            SpeedoXMLDescriptor desc = (SpeedoXMLDescriptor) itDesc.next();
            for (Iterator itPack = desc.jdoPackage.values().iterator(); itPack.hasNext();) {
                SpeedoPackage sp = (SpeedoPackage) itPack.next();
                for (Iterator itclass = sp.jdoClass.values().iterator(); itclass.hasNext();) {
                    SpeedoClass jdoclass = (SpeedoClass) itclass.next();

                    String name = sp.name + '.' + jdoclass.name;
                    String bindingName = sp.name + '.' + scp.mapperName + '.' +
                            NamingRules.bindingName(jdoclass.name);
                    String proxyName = NamingRules.proxyName(name);

                    ClassWriter cw = new ClassWriter(false);

                    RedundencyRemover rr = new RedundencyRemover(cw, logger);

                    List itfs = new ArrayList();

                    // merges the Binding class
                    Set superClasses = new TreeSet();
                    superClasses.add(name.replace('.', '/'));
                    ClassReader cr = loadJavaClass(false, bindingName, scp.output, true);
                    MergedClassModifier mmodifier =
                            new MergedClassModifier(rr, itfs, logger);
                    ClassRenamer crenamer = new ClassRenamer(mmodifier,
                            superClasses, name.replace('.', '/'), logger);
                    cr.accept(crenamer, false);

                    // merges the Proxy class
                    superClasses.add(bindingName.replace('.', '/'));
                    String superClass = jdoclass.persistenceCapableSuperClassName;
                    cr = loadJavaClass(false, proxyName, scp.output, true);
                    if (superClass == null || superClass.length() == 0) {
                        mmodifier = new MergedClassModifier(rr, itfs, logger);
                    } else {
                        mmodifier = new MergedClassModifier(rr, itfs,
                                superClass.replace('.', '/'), logger);
                    }
                    crenamer = new ClassRenamer(mmodifier, superClasses,
                            name.replace('.', '/'), logger);
                    cr.accept(crenamer, false);

                    // modifies the original code
                    cr = loadJavaClass(false, name, scp.output, false);
                    ClassAccessorModifier caccessmodifier =
                            new ClassAccessorModifier(rr, xmls, logger);
                    ClassInterfaceAdder citfadder = new ClassInterfaceAdder(
                            caccessmodifier, itfs, logger);

                    cr.accept(citfadder, false);
                    writeJavaClass(name, cw, scp.output);

                    if (jdoclass.identityType == SpeedoIdentity.USER_ID
                        && jdoclass.objectidClass != null) {
                        cw = new ClassWriter(false);
                        String cn = jdoclass.objectidClass;
                        cr = loadJavaClass(false, cn, scp.output, true);
                        cr.accept(new UserIdEnhancer(cw, jdoclass, logger), false);
                        writeJavaClass(cn, cw, scp.output);

                        cw = new ClassWriter(false);
                        cn = NamingRules.pnameName(jdoclass.objectidClass);
                        cr = loadJavaClass(false, cn, scp.output, true);
                        cr.accept(new PNameEnhancer(cw, jdoclass, logger), false);
                        writeJavaClass(cn, cw, scp.output);
                    }
                }
            }
        }

        // Displays thrown exceptions
        if (!except.isEmpty() && logger.isLoggable(BasicLevel.WARN)) {
            for (Iterator it = except.iterator(); it.hasNext();) {
                Exception e = (Exception) it.next();
                logger.log(BasicLevel.WARN, e.getLocalizedMessage());
            }
        }
    }
}
