/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.generation.enhancer;

import org.objectweb.asm.ClassAdapter;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.Constants;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Type;
import org.objectweb.asm.Label;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.metadata.SpeedoField;
import org.objectweb.speedo.metadata.SpeedoModifier;
import org.objectweb.speedo.generation.lib.NamingRules;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Iterator;
import java.util.Collection;

/**
 * Adds getter and setter methods to a class.
 *
 * Adapted from generateAccessors, genGetter and genSetter in EnhancerTool.
 */
public class ClassAccessorAdder extends LoggedClassAdapter {

    /**
     * The Speedo meta information for the visited class.
     */
    private final SpeedoClass jdoClass;

    /**
     * A collection of SpeedoXMLDescriptor describing known persistent classes
     */
    private final Collection xmlDescriptors;

    /**
     * Creates a new {@link ClassAccessorAdder}.
     *
     * @param cv the class visitor to be used to generate the modified class
     * @param jdoClass the Speedo meta information for the visited class.
     */
    public ClassAccessorAdder(final ClassVisitor cv,
                              final SpeedoClass jdoClass,
                              final Collection xmlDescriptors,
                              Logger logger) {
        super(cv, logger);
        this.jdoClass = jdoClass;
        this.xmlDescriptors = xmlDescriptors;
    }

    // IMPLEMENTATION OF THE ClassVisitor INTERFACE //
    // ---------------------------------------------//

    public void visit(final int access,
                      final String name,
                      final String superName,
                      final String[] interfaces,
                      final String sourceFile) {
        cv.visit(access, name, superName, interfaces, sourceFile);

        // Creates usable accessors for each jdo field
        Iterator i = jdoClass.jdoField.values().iterator();
        while (i.hasNext()) {
            SpeedoField jdoField = (SpeedoField) i.next();
            if (jdoField.name == null) {
                continue;
            }
            int acc = jdoField.access;
            if (jdoField.persistenceModifier >= SpeedoModifier.transactional) {
                // fields that are persistent or transactional
                if ((acc & Constants.ACC_STATIC) != 0) {
                    throw new RuntimeException(
                            "A static field cannot be defined as persistent");
                }
                genSetter(name, jdoField);
                if (jdoField.persistenceModifier == SpeedoModifier.persistent) {
                    genGetter(name, jdoField);
                }
            } else if ((acc &
                    (Constants.ACC_TRANSIENT | Constants.ACC_FINAL)) != 0) {
                // transient and/or final field
                genSetter(name, jdoField);
                genGetter(name, jdoField);
            } else if (jdoField.persistenceModifier == SpeedoModifier.missing) {
                if (Util.isPersistentType(jdoField.desc, xmlDescriptors)) {
                    // fields whose persistent-modifier isn't defined
                    genSetter(name, jdoField);
                    genGetter(name, jdoField);
                }
            }
        }
    }

    public CodeVisitor visitMethod(final int access,
                                   final String name,
                                   final String desc,
                                   final String[] exceptions) {
        // optimization: do not create an empty code adapter for visited
        // methods, as is done by default in ClassAdapter
        return cv.visitMethod(access, name, desc, exceptions);
    }

    // OTHER METHODS //
    // --------------//

    /**
     * Generates a getter method for the given field.
     *
     * @param owner the internal name of the visited class
     * @param jdoField the field for which the getter method must be generated
     */
    private void genGetter(final String owner, final SpeedoField jdoField) {
        int flags = jdoField.access;
        String name =
                NamingRules.getterName(jdoField.jdoClass, jdoField.name);
        if (debug) {
            logger.log(BasicLevel.DEBUG, "Generate the getter method '"
                    + name + "' on the class '" + jdoField.jdoClass.name + "'");
        }
        String desc = jdoField.desc;
        CodeVisitor cv = this.cv.visitMethod(flags, name, "()" + desc, null);
        Label next = new Label();
        // generate code corresponding to
        // 'SpeedoAccessor sa = getSpeedoAccessor();'
        cv.visitVarInsn(Constants.ALOAD, 0);
        cv.visitMethodInsn(
                Constants.INVOKEINTERFACE,
                "org/objectweb/speedo/api/SpeedoProxy",
                "getSpeedoAccessor",
                "()Lorg/objectweb/speedo/api/SpeedoAccessor;");
        cv.visitVarInsn(Constants.ASTORE, 1);
        // generate code to call readIntention
        cv.visitVarInsn(Constants.ALOAD, 0);
        cv.visitVarInsn(Constants.ALOAD, 1);
        int idx = jdoField.number / 64;
        String decl = "(Lorg/objectweb/speedo/api/SpeedoAccessor;J";
        for(int i=0; i<idx; i++) {
            decl += ";J";
            Util.visitLongConstant(cv, 0);
        }
        Util.visitLongConstant(cv, 1L << (jdoField.number % 64));
        cv.visitMethodInsn(
                Constants.INVOKEINTERFACE,
                "org/objectweb/speedo/api/SpeedoProxy",
                "jdoReadIntention",
                decl + ")V");
        // generate code corresponding to 'if (sa != this)'
        cv.visitVarInsn(Constants.ALOAD, 1);
        cv.visitVarInsn(Constants.ALOAD, 0);
        cv.visitJumpInsn(Constants.IF_ACMPEQ, next);
        // generate code to return optimistic field
        String fieldsOwner = NamingRules.fieldsName(owner);
        cv.visitVarInsn(Constants.ALOAD, 1);
        cv.visitTypeInsn(Constants.CHECKCAST, fieldsOwner);
        cv.visitFieldInsn(Constants.GETFIELD, fieldsOwner, jdoField.name, desc);
        Type returnType = Type.getType(desc);
        cv.visitInsn(returnType.getOpcode(Constants.IRETURN));
        // generate code to return 'this' field
        cv.visitLabel(next);
        cv.visitVarInsn(Constants.ALOAD, 0);
        cv.visitFieldInsn(Constants.GETFIELD, owner, jdoField.name, desc);
        cv.visitInsn(returnType.getOpcode(Constants.IRETURN));
        cv.visitMaxs(4, 2);
        jdoField.hasGetter = true;
    }

    /**
     * Generates a setter method for the given field.
     *
     * @param owner the internal name of the visited class
     * @param jdoField the field for which the setter method must be generated
     */
    private void genSetter(final String owner, final SpeedoField jdoField) {
        int flags = jdoField.access;
        String name =
                NamingRules.setterName(jdoField.jdoClass, jdoField.name);
        if (debug) {
            logger.log(BasicLevel.DEBUG, "Generate the setter method '"
                    + name + "' on the class '" + jdoField.jdoClass.name + "'");
        }
        String desc = jdoField.desc;
        CodeVisitor cv =
                this.cv.visitMethod(flags, name, "(" + desc + ")V", null);
        Type type = Type.getType(desc);
        int params = type.getSize() + 1;
        Label next = new Label();
        Label end = new Label();
        // generate code corresponding to (only for reference fields)
        // 'jdoMakeFieldPersistent(<field number>, <value>)'
        if (desc.charAt(0) == 'L' && !desc.equals("Ljava/lang/String;")) {
            cv.visitVarInsn(Constants.ALOAD, 0);
            Util.visitIntConstant(cv, jdoField.number);
            cv.visitVarInsn(Constants.ALOAD, 1);
            cv.visitMethodInsn(
                    Constants.INVOKEINTERFACE,
                    "org/objectweb/speedo/api/SpeedoProxy",
                    "jdoMakeFieldPersistent",
                    "(ILorg/objectweb/speedo/api/SpeedoAccessor;)V");
        }
        // generate code corresponding to
        // 'SpeedoAccessor sa = getSpeedoAccessor();'
        cv.visitVarInsn(Constants.ALOAD, 0);
        cv.visitMethodInsn(
                Constants.INVOKEINTERFACE,
                "org/objectweb/speedo/api/SpeedoProxy",
                "getSpeedoAccessor",
                "()Lorg/objectweb/speedo/api/SpeedoAccessor;");
        cv.visitVarInsn(Constants.ASTORE, params);
        // generate code to call writeIntention
        cv.visitVarInsn(Constants.ALOAD, 0);
        cv.visitVarInsn(Constants.ALOAD, params);
        int idx = jdoField.number / 64;
        String decl = "(Lorg/objectweb/speedo/api/SpeedoAccessor;J";
        for(int i=0; i<idx; i++) {
            decl += ";J";
            Util.visitLongConstant(cv, 0);
        }
        Util.visitLongConstant(cv, 1L << (jdoField.number % 64));
        cv.visitMethodInsn(
                Constants.INVOKEINTERFACE,
                "org/objectweb/speedo/api/SpeedoProxy",
                "jdoWriteIntention",
                decl + ")V");
        // generate code corresponding to 'if (sa != this)'
        cv.visitVarInsn(Constants.ALOAD, params);
        cv.visitVarInsn(Constants.ALOAD, 0);
        cv.visitJumpInsn(Constants.IF_ACMPEQ, next);
        // generate code to set optimistic field
        String fieldsOwner = NamingRules.fieldsName(owner);
        cv.visitVarInsn(Constants.ALOAD, params);
        cv.visitTypeInsn(Constants.CHECKCAST, fieldsOwner);
        cv.visitVarInsn(type.getOpcode(Constants.ILOAD), 1);
        cv.visitFieldInsn(Constants.PUTFIELD, fieldsOwner, jdoField.name, desc);
        cv.visitJumpInsn(Constants.GOTO, end);
        // generate code to set 'this' field
        cv.visitLabel(next);
        cv.visitVarInsn(Constants.ALOAD, 0);
        cv.visitVarInsn(type.getOpcode(Constants.ILOAD), 1);
        cv.visitFieldInsn(Constants.PUTFIELD, owner, jdoField.name, desc);
        // return
        cv.visitLabel(end);
        cv.visitInsn(Constants.RETURN);
        cv.visitMaxs(4, params + 1);
        jdoField.hasSetter = true;
    }
}
