/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.generation.enhancer;

import org.objectweb.asm.CodeAdapter;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Constants;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.ClassAdapter;
import org.objectweb.asm.Label;
import org.objectweb.asm.Type;
import org.objectweb.speedo.metadata.SpeedoField;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.generation.lib.NamingRules;
import org.objectweb.util.monolog.api.Logger;

import java.util.List;
import java.util.ArrayList;
import java.util.Collection;

/**
 * Replaces field accesses by calls to getter and setter methods.
 *
 * Adapted from modifyMethods and replaceInstruction in EnhancerTool.
 */
public class ClassAccessorModifier extends LoggedClassAdapter {

    /**
     * Stack size variation corresponding to each JVM instruction. This stack
     * variation is equal to the size of the values produced by an instruction,
     * minus the size of the values consumed by this instruction.
     */

    private final static int[] SIZE;

    /**
     * Computes the stack size variation corresponding to each JVM instruction.
     * Copied from the org.objectweb.asm.CodeWriter class.
     */

    static {
        int i;
        int[] b = new int[202];
        String s =
                "EFFFFFFFFGGFFFGGFFFEEFGFGFEEEEEEEEEEEEEEEEEEEEDEDEDDDDDCDCDEEEEEEEEE" +
                "EEEEEEEEEEEBABABBBBDCFFFGGGEDCDCDCDCDCDCDCDCDCDCEEEEDDDDDDDCDCDCEFEF" +
                "DDEEFFDEDEEEBDDBBDDDDDDCCCCCCCCEFEDDDCDCDEEEEEEEEEEFEEEEEEDDEEDDEE";
        for (i = 0; i < b.length; ++i) {
            b[i] = s.charAt(i) - 'E';
        }
        SIZE = b;
    }

    /**
     * A collections of SpeedoXMLDescriptors describing known persistent classes.
     */
    private final Collection xmlDescriptors;

    /**
     * Internal name of the visited class.
     */
    String owner;

    /**
     * Internal name of the corresponding "XXXFields" class.
     */
    String fieldsOwner;

    int nbfields = 128;

    /**
     * Constructs a new {@link ClassAccessorModifier}.
     *
     * @param cv the class visitor to be used to generate the modified class
     * @param xmlDescriptors the JDO Model describing known JDO Instances.
     */
    public ClassAccessorModifier(final ClassVisitor cv,
                                 final Collection xmlDescriptors,
                                 Logger logger) {
        super(cv, logger);
        this.xmlDescriptors = xmlDescriptors;
    }

    // IMPLEMENTATION OF THE ClassVisitor INTERFACE //
    // ---------------------------------------------//

    public void visit(final int access,
                      final String name,
                      final String superName,
                      final String[] interfaces,
                      final String sourceFile) {
        cv.visit(access, name, superName, interfaces, sourceFile);
        owner = name;
        nbfields = Util.isPersistentCapable(owner.replace('/', '.'), xmlDescriptors).computeFieldNumbers();
        fieldsOwner = NamingRules.fieldsName(owner);
    }

    public CodeVisitor visitMethod(final int access,
                                   final String name,
                                   final String desc,
                                   final String[] exceptions) {
        CodeVisitor cv = this.cv.visitMethod(access, name, desc, exceptions);
        if ((access & Constants.ACC_ABSTRACT) != 0) {
            // nothing to be modified for abstract methods
            return cv;
        }
        if (name.startsWith("jdo")) {
            // do not modify the accessor methods generated by the enhancer
            return cv;
        }
        if (name.equals("<init>") ||
                name.equals("<clinit>") ||
                (access & Constants.ACC_STATIC) != 0) {
            // for constructors, class initializer ansd static methods, do not
            // modify accesses to 'this' fields, but replace accesses to other
            // fields by calls to accessor methods
            return new DefaultCodeAccessorModifier(desc, false, cv);
        } else {
            // for other methods, generates two copies of the method, one for
            // pessimistic case, the other for the optimistic case

            // computes the access flags of the optimistic method
            int newAccess = access;
            newAccess &= ~(Constants.ACC_PUBLIC | Constants.ACC_PROTECTED);
            newAccess |= Constants.ACC_PRIVATE;

            // computes the descriptor of the optimistic method,
            // which has one additional formal parameter of type XXXFields
            Type returnType = Type.getReturnType(desc);
            Type[] types = Type.getArgumentTypes(desc);
            Type[] newTypes = new Type[types.length + 1];
            System.arraycopy(types, 0, newTypes, 0, types.length);
            newTypes[types.length] = Type.getType("L" + fieldsOwner + ";");
            String newDesc = Type.getMethodDescriptor(returnType, newTypes);

            // creates a code visitor to generate the optimistic method
            CodeVisitor newCv =
                    this.cv.visitMethod(newAccess, name, newDesc, exceptions);

            // wraps the code visitors to visit the pessimistic and optimistic
            // methods inside code modifiers, which will replace the field
            // accesses of the original method.
            // For the pessimistic method, fields accesses to 'this' are not
            // modified, but a header is added at the begining of the method
            // to call, if needed, the optimistic method (and to call read or
            // writeIntention otherwise
            // For the optimisitic method, fields accesses to 'this' are
            // replaced by field accesses to the state object (the added actual
            // parameter), and a header is added to call read or writeIntention
            cv = new NormalCodeAccessorModifier(desc, name, newDesc, cv);
            newCv = new OptimisticCodeAccessorModifier(desc, newCv);
            // creates a code duplicator with the two previous wrapped code
            // visitors: when the original method code will be visited, it will
            // then be duplicated, and modified by the previous wrappers
            return new CodeDuplicator(new CodeVisitor[]{cv, newCv});
        }
    }

    // OTHER METHODS //
    // --------------//

    /**
     * Looks for a specific <code>SpeedoField</code> in the object model.
     *
     * @param name the name of the field to be fetched
     * @param className the complete class name that the field belongs to
     * @return the corresponding SpeedoField if it exists, null either
     */
    SpeedoField fetchJDOField(final String name, final String className) {
        SpeedoClass c = Util.isPersistentCapable(className, xmlDescriptors);
        if (c != null) {
            return (SpeedoField) c.jdoField.get(name);
        } else {
            return null;
        }
    }

    /**
     * The default method code modifier, used for constructors, static methods
     * and class initializers. This code modifier does not modify accesses to
     * 'this' fields, but replaces accesses to other fields by calls to accessor
     * methods.
     */
    class DefaultCodeAccessorModifier extends CodeAdapter {

        /**
         * Descriptor of the visited method
         */
        protected String desc;

        /**
         * Types of the parameters of the visited method
         */
        protected Type[] types;

        /**
         * Total size of the formal parameters of the visited method
         */
        protected int params;

        /**
         * A label that designates the start of the original method code
         */
        private Label start;

        /**
         * A label that designates the start of the header added to the method
         */
        private Label header;

        /**
         * A list of instructions that have been visited but not yet regenerated
         * through {@link #cv cv}. This list is a list of Insn objects. It is
         * used to be able to modify an ALOAD 0 instruction long after it has
         * been visited by this visitor, when a GETFIELD or PUTFIELD instruction
         * is encountered.
         */
        protected List insns;

        /**
         * Symbolic state of the execution stack at the current bytecode
         * instruction. The last element corresponds to the top of the stack.
         * Each element is either <tt>null</tt> or an Integer object. The last
         * case signifies that, at execution time, and at the current bytecode
         * instruction, this stack element will contain 'this'. Moreover the
         * integer value indicates the index in 'insns' of the instruction that
         * pushed this value on the stack.
         */
        protected List stack;

        /**
         * Indicates the fields that the visited method may read on 'this'.
         * If the i-th bit is set to 1, then the i-th field may be read.
         */
        protected long[] readFields;

        /**
         * Indicates the fields that the visited method may write on 'this'.
         * If the i-th bit is set to 1, then the i-th field may be witten.
         */
        protected long[] writtenFields;

        /**
         * Constructs a new {@link DefaultCodeAccessorModifier}.
         *
         * @param desc the descriptor of the visited method
         * @param hasHeader indicates if a header must be added to the visited
         * method
         * @param cv the code visitor to be used to generate the modified method
         */
        public DefaultCodeAccessorModifier(final String desc,
                                           final boolean hasHeader,
                                           final CodeVisitor cv) {
            super(cv);
            int size = (nbfields/64) + ((nbfields % 64) > 0 ? 1 : 0);
            writtenFields = new long[size];
            readFields = new long[size];
            this.desc = desc;
            this.types = Type.getArgumentTypes(desc);
            params = 1;
            for (int i = 0; i < types.length; ++i) {
                params += types[i].getSize();
            }
            this.insns = new ArrayList();
            this.stack = new ArrayList();
            if (hasHeader) {
                start = new Label();
                header = new Label();
                cv.visitJumpInsn(Constants.GOTO, header);
                visitLabel(start);
            }
        }

        // IMPLEMENTATION OF THE CodeVisitor INTERFACE //
        // --------------------------------------------//

        public void visitInsn(final int opcode) {
            if ((opcode >= Constants.IRETURN && opcode <= Constants.RETURN) ||
                    opcode == Constants.ATHROW) {
                flushInsns();
                cv.visitInsn(opcode);
            } else {
                if (opcode == Constants.DUP && stack.size() > 0) {
                    // optimization for DUP
                    stack.add(stack.get(stack.size() - 1));
                } else {
                    updateStack(SIZE[opcode]);
                }
                insns.add(new Insn(opcode));
            }
        }

        public void visitIntInsn(final int opcode, final int operand) {
            updateStack(SIZE[opcode]);
            insns.add(new Insn(true, opcode, operand));
        }

        public void visitVarInsn(final int opcode, final int var) {
            if (opcode == Constants.RET) {
                flushInsns();
                cv.visitVarInsn(opcode, var);
            } else {
                if (opcode == Constants.ALOAD && var == 0) {
                    stack.add(new Integer(insns.size()));
                } else {
                    updateStack(SIZE[opcode]);
                }
                insns.add(new Insn(false, opcode, var));
            }
        }

        public void visitTypeInsn(final int opcode, final String desc) {
            updateStack(SIZE[opcode]);
            insns.add(new Insn(opcode, desc));
        }

        public void visitFieldInsn(final int opcode,
                                   final String owner,
                                   final String name,
                                   final String desc) {
            int size;
            // computes the stack size variation
            char c = desc.charAt(0);
            switch (opcode) {
            case Constants.GETSTATIC:
                size = c == 'D' || c == 'J' ? 2 : 1;
                break;
            case Constants.PUTSTATIC:
                size = c == 'D' || c == 'J' ? -2 : -1;
                break;
            case Constants.GETFIELD:
                size = c == 'D' || c == 'J' ? 1 : 0;
                break;
                //case Constants.PUTFIELD:
            default:
                size = c == 'D' || c == 'J' ? -3 : -2;
                break;
            }
            if (opcode == Constants.PUTFIELD) {
                SpeedoField jdofield =
                        fetchJDOField(name, owner.replace('/', '.'));
                if (jdofield != null && jdofield.hasSetter) {
                    int pushThisInsnIndex = -1;
                    if (c != 'L' || start == null) {
                        // putfield insns for non primitive fields are always
                        // replaced with calls to a setter method, which will
                        // make the field's value persistent, if needed.
                        int stackSize = stack.size();
                        if (stackSize >= 2) {
                            Integer i = (Integer) stack.get(stackSize - 2);
                            if (i != null) {
                                pushThisInsnIndex = i.intValue();
                            }
                        }
                    }
                    if (pushThisInsnIndex != -1) {
                        writtenFields[jdofield.number / 64] |= (1L << (jdofield.number % 64));
                        visitThisFieldInsn(
                                opcode, owner, name, desc, pushThisInsnIndex);
                    } else {
                        String setterName = NamingRules.setterName(
                                jdofield.jdoClass, jdofield.name);
                        String setterDesc = "(" + desc + ")V";
                        insns.add(new Insn(Constants.INVOKEVIRTUAL,
                                owner, setterName, setterDesc));
                    }
                    updateStack(size);
                    return;
                }
            } else if (opcode == Constants.GETFIELD) {
                SpeedoField jdofield =
                        fetchJDOField(name, owner.replace('/', '.'));
                if (jdofield != null && jdofield.hasGetter) {
                    int pushThisInsnIndex = -1;
                    int stackSize = stack.size();
                    if (stackSize >= 1) {
                        Integer i = (Integer) stack.get(stackSize - 1);
                        if (i != null) {
                            pushThisInsnIndex = i.intValue();
                        }
                    }
                    if (pushThisInsnIndex != -1) {
                        readFields[jdofield.number / 64] |= (1L << (jdofield.number % 64));
                        visitThisFieldInsn(
                                opcode, owner, name, desc, pushThisInsnIndex);
                    } else {
                        String getterName =
                                NamingRules.getterName(jdofield.jdoClass,
                                        jdofield.name);
                        String getterDesc = "()" + desc;
                        insns.add(new Insn(Constants.INVOKEVIRTUAL,
                                owner, getterName, getterDesc));
                    }
                    updateStack(size);
                    return;
                }
            }
            updateStack(size);
            insns.add(new Insn(opcode, owner, name, desc));
        }

        public void visitMethodInsn(final int opcode,
                                    final String owner,
                                    final String name,
                                    final String desc) {
            // computes the stack size variation
            int size = opcode == Constants.INVOKESTATIC ? 0 : 1;
            int c = 1;
            while (true) {
                char car = desc.charAt(c++);
                if (car == ')') {
                    car = desc.charAt(c);
                    if (car == 'V') {
                        size = -size;
                    } else {
                        size = (car == 'D' || car == 'J' ? 2 : 1) - size;
                    }
                    break;
                } else if (car == 'L') {
                    while (desc.charAt(c++) != ';') {
                    }
                    size += 1;
                } else if (car == '[') {
                    while ((car = desc.charAt(c)) == '[') {
                        ++c;
                    }
                    if (car == 'D' || car == 'J') {
                        size -= 1;
                    }
                } else if (car == 'D' || car == 'J') {
                    size += 2;
                } else {
                    size += 1;
                }
            }
            updateStack(size);
            insns.add(new Insn(opcode, owner, name, desc));
        }

        public void visitJumpInsn(final int opcode, final Label label) {
            flushInsns();
            cv.visitJumpInsn(opcode, label);
        }

        public void visitLabel(final Label label) {
            insns.add(new Insn(label));
        }

        public void visitLdcInsn(final Object cst) {
            int size;
            if (cst instanceof Double || cst instanceof Long) {
                size = 2;
            } else {
                size = 1;
            }
            updateStack(size);
            insns.add(new Insn(cst));
        }

        public void visitIincInsn(final int var, final int increment) {
            // no stack change
            insns.add(new Insn(var, increment));
        }

        public void visitTableSwitchInsn(final int min,
                                         final int max,
                                         final Label dflt,
                                         final Label labels[]) {
            flushInsns();
            cv.visitTableSwitchInsn(min, max, dflt, labels);
        }

        public void visitLookupSwitchInsn(final Label dflt,
                                          final int keys[],
                                          final Label labels[]) {
            flushInsns();
            cv.visitLookupSwitchInsn(dflt, keys, labels);
        }

        public void visitMultiANewArrayInsn(final String desc,
                                            final int dims) {
            updateStack(1 - dims);
            insns.add(new Insn(desc, dims));
        }

        public void visitTryCatchBlock(final Label start,
                                       final Label end,
                                       final Label handler,
                                       final String type) {
            flushInsns();
            super.visitTryCatchBlock(start, end, handler, type);
        }

        public void visitMaxs(final int maxStack, final int maxLocals) {
            flushInsns();
            if (header != null) {
                cv.visitLabel(header);
                generateMethodHeader();
                cv.visitJumpInsn(Constants.GOTO, start);
            }
            cv.visitMaxs(maxStack, maxLocals);
        }

        public void visitLocalVariable(final String name,
                                       final String desc,
                                       final Label start,
                                       final Label end,
                                       final int index) {
            flushInsns();
            super.visitLocalVariable(name, desc, start, end, index);
        }

        public void visitLineNumber(final int line, final Label start) {
            flushInsns();
            super.visitLineNumber(line, start);
        }

        // OTHER METHODS //
        // --------------//

        /**
         * Generates a header which is added at the begining of the the visited
         * method. The default implementation of this method does nothing.
         */
        protected void generateMethodHeader() {
            // does nothing
        }

        /**
         * Visits field access instruction on 'this'.
         *
         * @param opcode GETFIELD or PUTFIELD
         * @param owner internal name of the owner class of the field
         * @param name name of the field
         * @param desc type descriptor of te field
         * @param pushThisInsn index in {@link #insns insns} of the ALOAD 0
         * instruction that pushed the 'this' value which is, at this bytecode
         * instruction, at the top (for GETFIELD) or just under the top (for
         * PUTFIELD) of the stack.
         */
        protected void visitThisFieldInsn(final int opcode,
                                          final String owner,
                                          final String name,
                                          final String desc,
                                          final int pushThisInsn) {
            // keeps the field access instructions on 'this' unchanged
            insns.add(new Insn(opcode, owner, name, desc));
        }

        /**
         * Makes {@link #cv cv} visit the instructions stored in {@link #insns
         * insns}, and then clear this list, as well as the stack. This method
         * is used when the end of method is encountered, but also each time
         * a jump or return statement is visited, because we can not easily know
         * the stack state at the next instruction (this is possible but
         * complicated: we prefer to keep the algorithm simple. The cost of this
         * simplification is that some field accesses on 'this' will not be
         * statically detected, and will therefore be replaced by calls to
         * accessor methods, which is less efficient but safe).
         */
        protected void flushInsns() {
            int n = insns.size();
            for (int i = 0; i < n; ++i) {
                ((Insn) insns.get(i)).accept(cv);
            }
            insns.clear();
            stack.clear();
        }

        /**
         * Pushes some <tt>null</tt> elements on the stack, or pops some
         * elements.
         * @param n the number of elements to be pushed (if n is positive), or
         * to be popped (if n is negative).
         */
        private void updateStack(final int n) {
            int size = stack.size();
            if (n > 0) {
                for (int i = 0; i < n; ++i) {
                    stack.add(null);
                }
            } else {
                int m = -n;
                if (m >= size) {
                    stack.clear();
                } else {
                    for (int i = 0; i < m; ++i) {
                        stack.remove(--size);
                    }
                }
            }
        }
    }

    /**
     * The method code modifier used for pessimistic methods. This code modifier
     * does not modify accesses to 'this' fields, but replaces accesses to other
     * fields by calls to accessor methods. It also adds a header to the method
     * to call the optimistic method, if needed, and to call read or
     * writeIntention.
     */
    class NormalCodeAccessorModifier extends DefaultCodeAccessorModifier {

        /**
         * Name of the optimistic method.
         */
        protected String optimisticName;

        /**
         * Descriptor of the corresponding optimistic method.
         */
        protected String optimisticDesc;

        /**
         * Constructs a new {@link NormalCodeAccessorModifier}.
         *
         * @param desc the descriptor of the pessimistic method
         * @param optimisticName the name of the optimisitic method
         * @param optimisticDesc the descriptor of the optimisitic method
         * @param cv the code visitor to be used to generate the modified method
         */
        public NormalCodeAccessorModifier(final String desc,
                                          final String optimisticName,
                                          final String optimisticDesc,
                                          final CodeVisitor cv) {
            super(desc, true, cv);
            this.optimisticName = optimisticName;
            this.optimisticDesc = optimisticDesc;
        }

        // OVERRIDEN METHODS //
        // ------------------//

        public void visitMaxs(final int maxStack, final int maxLocals) {
            boolean allUnmodified = true;
            for (int i=0; allUnmodified && i<nbfields; i++) {
                allUnmodified &= readFields[i/64] == 0 & writtenFields[i/64] == 0;
            }
            if (allUnmodified) {
                super.visitMaxs(maxStack, maxLocals);
            } else {
                super.visitMaxs(
                        Math.max(maxStack, Math.max(params + (nbfields/64) + 1, 4)),
                        Math.max(maxLocals, params + (nbfields/64) + 1));
            }
        }

        protected void generateMethodHeader() {
            boolean allUnmodified = true;
            boolean hasWritten = false;
            for (int i=0; allUnmodified && i<nbfields; i++) {
                hasWritten |= writtenFields[i/64] != 0;
                allUnmodified &= readFields[i/64] == 0 & writtenFields[i/64] == 0;
            }
            if (allUnmodified) {
                return;
            }
            Label next = new Label();
            // generate code corresponding to
            // 'SpeedoAccessor sa = getSpeedoAccessor();'
            cv.visitVarInsn(Constants.ALOAD, 0);
            cv.visitMethodInsn(
                    Constants.INVOKEINTERFACE,
                    "org/objectweb/speedo/api/SpeedoProxy",
                    "getSpeedoAccessor",
                    "()Lorg/objectweb/speedo/api/SpeedoAccessor;");
            cv.visitVarInsn(Constants.ASTORE, params);
            // generate code corresponding to 'if (sa != this)'
            cv.visitVarInsn(Constants.ALOAD, params);
            cv.visitVarInsn(Constants.ALOAD, 0);
            cv.visitJumpInsn(Constants.IF_ACMPEQ, next);
            // generate code to call optimistic method
            int var = 1;
            cv.visitVarInsn(Constants.ALOAD, 0);
            for (int i = 0; i < types.length; ++i) {
                cv.visitVarInsn(types[i].getOpcode(Constants.ILOAD), var);
                var += types[i].getSize();
            }
            cv.visitVarInsn(Constants.ALOAD, params);
            cv.visitTypeInsn(Constants.CHECKCAST, fieldsOwner);
            cv.visitMethodInsn(
                    Constants.INVOKESPECIAL,
                    owner,
                    optimisticName,
                    optimisticDesc);
            Type returnType = Type.getReturnType(desc);
            if (returnType == Type.VOID_TYPE) {
                cv.visitInsn(Constants.RETURN);
            } else {
                cv.visitInsn(returnType.getOpcode(Constants.IRETURN));
            }
            // generate code to call read or writeIntention
            cv.visitLabel(next);
            cv.visitVarInsn(Constants.ALOAD, 0);
            cv.visitVarInsn(Constants.ALOAD, 0);
            String decl = "(Lorg/objectweb/speedo/api/SpeedoAccessor";
            for(int i=0; i<readFields.length; i++) {
                Util.visitLongConstant(cv, readFields[i]);
                decl += ";J";
            }
            cv.visitMethodInsn(
                    Constants.INVOKEINTERFACE,
                    "org/objectweb/speedo/api/SpeedoProxy",
                    hasWritten ? "jdoWriteIntention" : "jdoReadIntention",
                    decl + ")V");
        }
    }

    /**
     * The method code modifier used for optimistic methods. This code modifier
     * replaces accesses to 'this' fields by accesses to the 'state' object, and
     * replaces accesses to other fields by calls to accessor methods. It also
     * adds a header to the method to call read or writeIntention.
     */
    class OptimisticCodeAccessorModifier extends DefaultCodeAccessorModifier {

        /**
         * Constructs a new {@link OptimisticCodeAccessorModifier}.
         *
         * @param desc the descriptor of the PESSIMISTIC method, and not of
         * the optimistic one
         * @param cv the code visitor to be used to generate the modified method
         */
        public OptimisticCodeAccessorModifier(final String desc,
                                              final CodeVisitor cv) {
            super(desc, true, cv);
        }

        // OVERRIDEN METHODS //
        // ------------------//

        public void visitVarInsn(final int opcode, final int var) {
            // shifts local variables that are just after the original actual
            // parameters (ie without counting the added formal paramter), to
            // make room for the added actual parameter
            super.visitVarInsn(opcode, var >= params ? var + 1 : var);
        }

        public void visitIincInsn(final int var, final int increment) {
            // shifts local variables that are just after the original actual
            // parameters (ie without counting the added formal paramter), to
            // make room for the added actual parameter
            super.visitIincInsn(var >= params ? var + 1 : var, increment);
        }

        public void visitMaxs(final int maxStack, final int maxLocals) {
            boolean allUnmodified = true;
            for (int i=0; allUnmodified && i<nbfields; i++) {
                allUnmodified &= readFields[i/64] == 0 & writtenFields[i/64] == 0;
            }
            if (allUnmodified) {
                super.visitMaxs(maxStack, maxLocals + (nbfields/64) + 1);
            } else {
                super.visitMaxs(Math.max(maxStack, 4), maxLocals + (nbfields/64) + 1);
            }
        }

        protected void generateMethodHeader() {
            boolean allUnmodified = true;
            boolean hasWritten = false;
            for (int i=0; allUnmodified && i<nbfields; i++) {
                hasWritten |= writtenFields[i/64] != 0;
                allUnmodified &= readFields[i/64] == 0 & writtenFields[i/64] == 0;
            }
            if (allUnmodified) {
                return;
            }
            // generates code to call the read or writeIntention method
            cv.visitVarInsn(Constants.ALOAD, 0);
            cv.visitVarInsn(Constants.ALOAD, params);
            String decl = "(Lorg/objectweb/speedo/api/SpeedoAccessor";
            for(int i=0; i<readFields.length; i++) {
                Util.visitLongConstant(cv, readFields[i]);
                decl += ";J";
            }
            cv.visitMethodInsn(
                    Constants.INVOKEINTERFACE,
                    "org/objectweb/speedo/api/SpeedoProxy",
                    hasWritten ? "jdoWriteIntention" : "jdoReadIntention",
                    decl + ")V");
        }

        protected void visitThisFieldInsn(final int opcode,
                                          final String owner,
                                          final String name,
                                          final String desc,
                                          final int pushThisInsn) {
            // replaces the field access instructions on 'this' by field
            // access instructions on the state object
            insns.set(pushThisInsn, new Insn(false, Constants.ALOAD, params));
            insns.add(new Insn(opcode, fieldsOwner, name, desc));
        }
    }

    /**
     * Object representation of a sequential bytecode instruction. A sequential
     * bytecode instruction is any instruction that is not a jump, a return,
     * a ret or a switch instruction. A single class is used to represent
     * several types of bytecode instruction, in order to minimize the size of
     * the outer class.
     */
    static class Insn {

        // the sequential bytecode instruction types
        final static int INSN = 1;
        final static int INT_INSN = 2;
        final static int VAR_INSN = 3;
        final static int TYPE_INSN = 4;
        final static int FIELD_INSN = 5;
        final static int METH_INSN = 6;
        final static int LABEL_INSN = 7;
        final static int LDC_INSN = 8;
        final static int IINC_INSN = 9;
        final static int MANA_INSN = 10;

        /**
         * Type of this instruction
         */
        private int type;

        /**
         * Opcode of this instruction
         */
        private int opcode;

        // operand(s) of this instruction

        private int int1, int2;
        private String str1, str2, str3;
        private Object obj;

        public Insn(final int opcode) {
            this.type = INSN;
            this.opcode = opcode;
        }

        public Insn(final boolean intInsn, final int opcode, final int val) {
            this.type = intInsn ? INT_INSN : VAR_INSN;
            this.opcode = opcode;
            this.int1 = val;
        }

        public Insn(final int opcode, final String desc) {
            this.type = TYPE_INSN;
            this.opcode = opcode;
            this.str1 = desc;
        }

        public Insn(final int opcode,
                    final String owner,
                    final String name,
                    final String desc) {
            this.type = opcode <= Constants.PUTFIELD ? FIELD_INSN : METH_INSN;
            this.opcode = opcode;
            this.str1 = owner;
            this.str2 = name;
            this.str3 = desc;
        }

        public Insn(final Label label) {
            this.type = LABEL_INSN;
            this.obj = label;
        }

        public Insn(final Object cst) {
            this.type = LDC_INSN;
            this.obj = cst;
        }

        public Insn(final int var, final int inc) {
            this.type = IINC_INSN;
            this.int1 = var;
            this.int2 = inc;
        }

        public Insn(final String desc, final int dims) {
            this.type = MANA_INSN;
            this.str1 = desc;
            this.int1 = dims;
        }

        /**
         * Makes the given code visitor visit this instruction.
         *
         * @param cv a code visitor
         */
        public void accept(final CodeVisitor cv) {
            switch (type) {
            case INSN:
                cv.visitInsn(opcode);
                break;
            case INT_INSN:
                cv.visitIntInsn(opcode, int1);
                break;
            case VAR_INSN:
                cv.visitVarInsn(opcode, int1);
                break;
            case TYPE_INSN:
                cv.visitTypeInsn(opcode, str1);
                break;
            case FIELD_INSN:
                cv.visitFieldInsn(opcode, str1, str2, str3);
                break;
            case METH_INSN:
                cv.visitMethodInsn(opcode, str1, str2, str3);
                break;
            case LABEL_INSN:
                cv.visitLabel((Label) obj);
                break;
            case LDC_INSN:
                cv.visitLdcInsn(obj);
                break;
            case IINC_INSN:
                cv.visitIincInsn(int1, int2);
                break;
            default:
                cv.visitMultiANewArrayInsn(str1, int1);
            }
        }
    }
}
