/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 0.1
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: P. Dechamboux, Y. Devaux, F. Dhumeaux.
 *
 */

package org.objectweb.speedo.generation.generator.proxy;

import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;
import org.apache.velocity.context.Context;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.api.SpeedoGenerationException;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.api.UserFieldMapping;
import org.objectweb.speedo.generation.generator.lib.AbstractVelocityGenerator;
import org.objectweb.speedo.generation.lib.NamingRules;
import org.objectweb.speedo.metadata.MetaDataHelper;
import org.objectweb.speedo.metadata.SpeedoArray;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.metadata.SpeedoField;
import org.objectweb.speedo.metadata.SpeedoIdentity;
import org.objectweb.speedo.metadata.SpeedoCollection;
import org.objectweb.speedo.metadata.SpeedoMap;
import org.objectweb.speedo.metadata.SpeedoExtension;
import org.objectweb.util.monolog.wrapper.velocity.VelocityLogger;
import org.objectweb.asm.Type;
import org.objectweb.jorm.comp.generator.lib.CommonHelper;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.Mapping;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassProject;
import org.objectweb.jorm.metainfo.api.CompositeName;
import org.objectweb.jorm.runtime.JormPathHelper;
import org.objectweb.jorm.naming.lib.CTHelper;
import org.objectweb.jorm.type.api.PType;

import java.io.FileWriter;
import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;
import java.util.List;

/**
 * This class is used to generate the file proxy.
 * This file will represent the persistent instance.
 * @author	F. Dhumeaux
 */
public class ProxyGenerator extends AbstractVelocityGenerator {

    public final static String LOGGER_NAME
            = "org.objectweb.speedo.generation.generator.proxy";

    public final static String TEMPLATE_NAME = TEMPLATE_DIR + ".proxy.Proxy";

    public final static String CID_BINDER_CLASS_NAME
            = "org.objectweb.jorm.facility.naming.polymorphid.PolymorphIdBinder";

    public final static String PNAME_SINGLE_BINDER_CLASS_NAME
            = "org.objectweb.jorm.facility.naming.basidir.BasidBinder";

    /**
     * is the name of the composite name used for the identifiers in case of
     * container identifier management.
     */
    public final static String SPEEDO_ID_NAME
            = "org.objectweb.jorm.facility.naming.polymorphid.PolymorphId";

    // IMPLEMENTATION OF THE GeneratorComponent INTERFACE //
    //----------------------------------------------------//

    public void init() throws SpeedoException {
        logger = scp.loggerFactory.getLogger(LOGGER_NAME);
    }


    // IMPLEMENTATION OF THE VelocityGenerator INTERFACE //
    //---------------------------------------------------//

    /**
     * This method generates the new file.
     *
     * @param sClass is the SpeedoClass
     * @param fileName name of the new file.
     * @exception SpeedoGenerationException if there is a problem during writing
     * the new file.
     */
    public void generate(SpeedoClass sClass, String fileName)
            throws SpeedoException {
        //EnhancerTool.logger = logger;
        if (templateFN == null)
            templateFN = TEMPLATE_NAME.replace('.', fs) + ".vm";
        try {
            if (!ve.templateExists(templateFN)) {
                templateFN = templateFN.replace(fs, '/');
                if (!ve.templateExists(templateFN)) {
                    throw new SpeedoException("The Speedo template " + templateFN
                        + " is not availlable in the classpath");
                }
            }
            Context ctx = getContext(sClass);
            FileWriter fw = new FileWriter(fileName);
            ve.getTemplate(templateFN).merge(ctx, fw);
            ve.setProperty(Velocity.RUNTIME_LOG_LOGSYSTEM, new VelocityLogger(logger));
            fw.flush();
            fw.close();
        } catch (Exception e) {
            throw new SpeedoGenerationException(
                    "Error during the generation of the file " + fileName, e);
        }
    }

    // PRIVATE METHODS //
    //-----------------//

    /**
     * This method initialises the Velocity context.
     *
     * @param jdoClass : the jdoClass which represents the class to generate.
     * @return a Velocity context.
     */
    private Context getContext(SpeedoClass jdoClass) throws SpeedoException {

        int nbField = jdoClass.computeFieldNumbers();
        //creation of the Velocity context
        Context ctx = new VelocityContext();

        ctx.put("jdoXmlDescFN", jdoClass.jdoPackage.jdoXMLDescriptor.xmlFile
                .replace(File.separatorChar, '.'));

        //class declaration
        ctx.put("signatureClass", jdoClass.signature);
        ctx.put("isAbstract", new Boolean(jdoClass.isAbstract));
        ctx.put("package", jdoClass.jdoPackage.name);
        ctx.put("mapper", scp.mapperName);
        ctx.put("projectName", scp.projectName);
        ctx.put("isSerializable", new Boolean(jdoClass.isSerializable));
        ctx.put("serialversionuid", new Long(jdoClass.VersionUID));
        if (jdoClass.identityType == SpeedoIdentity.CONTAINER_ID) {
            ctx.put("isContainerId", Boolean.TRUE);
            ctx.put("isContainerIdValue", "true");
        } else {
            ctx.put("isContainerId", Boolean.FALSE);
            ctx.put("isContainerIdValue", "false");
        }

        //class name
        ctx.put("baseClassName", jdoClass.name);
        ctx.put("className", NamingRules.proxyName(jdoClass.name));
        ctx.put("classNameBinding", NamingRules.bindingName(jdoClass.name));
        ctx.put("classNameFields", NamingRules.fieldsName(jdoClass.name));
        ctx.put("classNameKey", NamingRules.keyName(jdoClass.name));
        ctx.put("classNameAccessor", NamingRules.accessorName(jdoClass.name));

        //super class
        boolean hasSuperclass = !(jdoClass.persistenceCapableSuperClassName.equals(""));
        ctx.put("hasSuperclass", new Boolean(hasSuperclass));
        ctx.put("superclass", NamingRules.proxyName(jdoClass.persistenceCapableSuperClassName));


        // persistent fields
        StringBuffer fieldTypeArray = new StringBuffer("{");
        StringBuffer fieldNameArray = new StringBuffer("{");
        StringBuffer fieldFlagArray = new StringBuffer("{");
        List fields = new ArrayList(jdoClass.jdoField.size());
        List referenceFields = new ArrayList(jdoClass.jdoField.size());
        List primitiveFields = new ArrayList(jdoClass.jdoField.size());

        boolean hasRef = false;
        CommonHelper jormTools = new CommonHelper();
        String declLong = "";
        String useofLong = "";
        String sep = "";
        int ndlong = (jdoClass.computeFieldNumbers() / 64) + 1;
        for (int i=0; i<ndlong ; i++) {
            declLong += sep + " long fields" + i;
            useofLong += sep + " fields" + i;
            sep = ",";
        }
        ctx.put("declLong", declLong);
        ctx.put("useofLong", useofLong);
        for (Iterator it = jdoClass.jdoField.values().iterator(); it.hasNext();) {
            SpeedoField sp = (SpeedoField) it.next();
            Field f = new Field();
            f.name = sp.name;
            f.number = sp.number;
            f.getter = NamingRules.getterName(sp.jdoClass, sp.name);
            f.setter = NamingRules.setterName(sp.jdoClass, sp.name);
            f.isClassical = isClassicalType(sp.type());
            fields.add(f);
            f.modifier = sp.modifier();
            fieldTypeArray.append(sp.type());
            fieldTypeArray.append(".class, ");

            fieldNameArray.append("\"");
            fieldNameArray.append(sp.name);
            fieldNameArray.append("\", ");

            fieldFlagArray.append("READ_WRITE_OK, ");
            hasRef = hasRef || !f.isClassical;
            f.declaration = sp.privateSignature();
            f.isKey = sp.primaryKey;

            // looks for identity type of the field and tests if it is an array
            Type fieldType = Type.getReturnType(sp.desc);
            if (!f.isClassical && fieldType.getSort() == Type.OBJECT) {
                SpeedoClass jdoclass = MetaDataHelper.fetchJDOClass(
                        fieldType.getClassName(),
                        sp.jdoClass.jdoPackage.jdoXMLDescriptor);
                if (jdoclass != null) {
                    f.isContainerId = jdoclass.identityType == SpeedoIdentity.CONTAINER_ID;
                } else {
                    f.isContainerId = true;
                }
                f.isArray = false;
            } else {
                f.isArray = sp.jdoTuple instanceof SpeedoArray;
            }
            f.jormMeth = jormTools.upperFL(f.name);
            f.jormFieldIdLongPos = sp.number / 64;
            f.jormFieldId = 1L << (sp.number % 64);
            f.jormFielIdDecl = "";
            sep = "";
            for(int i=0; i<= (nbField/64); i++) {
                f.jormFielIdDecl += sep + (i == f.jormFieldIdLongPos
                        ? f.jormFieldId + "L" : "0L");
                sep = ", ";
            }
            TypedElement te = jdoClass.jormclass.getTypedElement(f.name);
            if (jormTools.isReference(te)) {
                f.isReference = true;
                referenceFields.add(f);
                f.jormType = "Object";
                f.linkedField = getLinkedField((Reference) te);
                f.isMultiValued = jormTools.isGenClassRef(te);
                if (!f.isMultiValued) {
                    f.memoryType = te.getType().getJormName();
                } else {
                    f.gcproxy = "org.objectweb.speedo.runtime.genclass.";
                    if (sp.jdoTuple instanceof SpeedoCollection) {
                        f.memoryType = "java.util.Collection";
                        f.gcproxy += "collection.CollectionImpl";
                    } else if (sp.jdoTuple instanceof SpeedoMap) {
                        f.memoryType = "java.util.Map";
                        f.gcproxy += "map.MapImpl";
                    } else {
                        f.memoryType = "Object";
                    }
                }
            } else {
                primitiveFields.add(f);
                f.jormType = te.getType().getJavaName();
                f.memoryType = sp.type();

                f.toMemory = "val";
                f.toStorage = f.name;
                SpeedoExtension se = sp.getExtensionByKey(
                        SpeedoProperties.FIELD_CONVERTER);
                if (se != null) {
                    f.userFieldMapping = se.value;
                    try {
                        UserFieldMapping ufm = (UserFieldMapping)
                                java.lang.Class.forName(se.value).newInstance();
                        if (ufm.getStorageType().isPrimitive()) {
                            f.toMemory = "new "
                                    + getJavaLangType(ufm.getStorageType())
                                    + "(" + f.toMemory + ")";
                        }
                        if (ufm.getMemoryType().isPrimitive()) {
                            f.toStorage = "new "
                                    + getJavaLangType(ufm.getMemoryType())
                                    + "(" + f.toStorage + ")";
                        }
                        f.toMemory = f.name + "UFM.toMemory(" + f.toMemory + ")";
                        f.toStorage = f.name + "UFM.toStorage(" + f.toStorage + ")";
                        if (ufm.getStorageType().isPrimitive()) {
                            f.toStorage = "(("
                                    + getJavaLangType(ufm.getStorageType())
                                    + ") " + f.toStorage + ")."
                                    + getValueGetter(ufm.getStorageType()) + "()";
                        } else {
                            f.toStorage = "("
                                    + ufm.getStorageType().getName()
                                    + ") " + f.toStorage;
                        }
                        if (ufm.getMemoryType().isPrimitive()) {
                            f.toMemory = "(("
                                    + getJavaLangType(ufm.getMemoryType())
                                    + ") " + f.toMemory + ")."
                                    + getValueGetter(ufm.getMemoryType()) + "()";
                        } else {
                            f.toMemory = "("
                                    + ufm.getMemoryType().getName()
                                    + ") " + f.toMemory;
                        }
                    } catch (Exception e) {
                        throw new SpeedoException(
                                "Impossible to instanciate the UserFieldMapping class '"
                                + se.value + "' for the field '" + sp.name
                                + "' of the class '" + jdoClass.getFQName() + "':", e);
                    }
                }
            }

        }
        ctx.put("fields", fields);
        ctx.put("primitiveFields", primitiveFields);
        ctx.put("referenceFields", referenceFields);
        ctx.put("hasRef", new Boolean(hasRef));
        ctx.put("fieldTypeArray", (fieldTypeArray.length() > 1
            ? fieldTypeArray.substring(0, fieldTypeArray.length() - 2) + "}"
            : fieldTypeArray.toString() + "}"));
        ctx.put("fieldNameArray", (fieldNameArray.length() > 1
            ? fieldNameArray.substring(0, fieldNameArray.length() - 2) + "}"
            : fieldNameArray.toString() + "}"));
        ctx.put("fieldFlagArray", (fieldFlagArray.length() > 1
            ? fieldFlagArray.substring(0, fieldFlagArray.length() - 2) + "}"
            : fieldFlagArray.toString() + "}"));

        // Jorm config
        ctx.put("jormconfig", getJormConfig(jdoClass));

        //Define the 'pnameHints' attribute.
        String pnameHints;
        if (jdoClass.identityType == SpeedoIdentity.USER_ID) {
            NameDef nd = getClassNameDef(jdoClass.jormclass);
            if (nd.isFieldName()) {
                //In case of the name def is a single field, provide the field value
                // as Object format
                String fn = nd.getFieldName();
                int tc = jdoClass.jormclass.getTypedElement(fn).getType().getTypeCode();
                switch (tc) {
                case PType.TYPECODE_CHAR:
                    pnameHints = "new Character(" + fn + ")";
                    break;
                case PType.TYPECODE_BYTE:
                    pnameHints = "new Byte(" + fn + ")";
                    break;
                case PType.TYPECODE_SHORT:
                    pnameHints = "new Short(" + fn + ")";
                    break;
                case PType.TYPECODE_INT:
                    pnameHints = "new Integer(" + fn + ")";
                    break;
                case PType.TYPECODE_LONG:
                    pnameHints = "new Long(" + fn + ")";
                    break;
                default:
                    pnameHints = fn;
                }
            } else {
                //In case of composite name the SpeedoProxy implements the PNG
                // interface.
                pnameHints = "this";
            }
        } else {
             pnameHints = "null";
        }
        ctx.put("pnameHints", pnameHints);

        return ctx;
    }

    public class Const {
        public String begin;
        public String end;

        public String getBegin() {
            return begin;
        }

        public String getEnd() {
            return end;
        }
    }

    public NameDef getRefNameDef(GenClassRef gcr,
                                 Class clazz) throws SpeedoException {
        return (NameDef) getMapping(clazz)
                .getGenClassMapping(gcr.getGenClassId())
                .getIdentifierMapping().getLinkedMO();
    }

    public NameDef getElemNameDef(GenClassRef gcr,
                                 Class clazz) throws SpeedoException {
        return (NameDef) getMapping(clazz)
                .getGenClassMapping(gcr.getGenClassId())
                .getReferenceMapping().getLinkedMO();
    }

    public NameDef getRefNameDef(ClassRef cr,
                                 Class clazz) throws SpeedoException {
        return (NameDef) getMapping(clazz).getClassMapping()
                .getReferenceMapping(cr.getName()).getLinkedMO();
    }

    private Mapping getMapping(Class clazz) throws SpeedoException {
        ClassProject cp = clazz.getClassProject(scp.projectName);
        if (cp == null) {
            throw new SpeedoException("No classproject found for the class "
                    + clazz.getFQName() + " and  the project " + scp.projectName);
        }
        int idx =  scp.mapperName.indexOf('.');
        Mapping m = cp.getMapping(idx == -1
                ? scp.mapperName
                : scp.mapperName.substring(0, idx));
        if (m == null) {
            throw new SpeedoException("No mapping found for the class "
                    + clazz.getFQName() + ",  the project " + scp.projectName
                    + " and  the mapper " + scp.mapperName);
        }
        return m;
    }

    private NameDef getClassNameDef(Class clazz) throws SpeedoException {
        return (NameDef) getMapping(clazz).getClassMapping()
                .getIdentifierMapping().getLinkedMO();
    }

    /**
     * This method transforms the first letter of a string into capital letter.
     *
     * @param var the String to transform.
     * @return the new String.
     */
    private String changeCasse(String var) {
        char[] tmp = var.toCharArray();
        tmp[0] = Character.toUpperCase(tmp[0]);
        return new String(tmp);
    }

    private String getLinkedField(Reference r) {
        if (r instanceof GenClassRef) {
            return JormPathHelper.getPath((GenClassRef) r, false);
        } else if (r instanceof ClassRef) {
            return JormPathHelper.getPath((ClassRef) r);
        } else {
            return null;
        }
    }

    private String getJormConfig(SpeedoClass jdoClass) throws SpeedoException {
        //Build the map
        Map jormconfig = new HashMap();

        // Put the binder class name of the current persistent class
        String id = JormPathHelper.getPath(jdoClass.jormclass);
        NameDef nd = getClassNameDef(jdoClass.jormclass);
        if (nd.isFieldName()) {
            jormconfig.put(id, PNAME_SINGLE_BINDER_CLASS_NAME
                + getCodingType(nd.getFieldName(), jdoClass.jormclass));
        } else if (nd.isNameRef()) {
            jormconfig.put(id, getBinderClassName(nd, false));
        }
        //Put the binder class name for each reference field
        for (Iterator it = jdoClass.jormclass.iterateAllField(); it.hasNext();) {
            Object o = it.next();
            if (o instanceof ClassRef) {
                nd = getRefNameDef((ClassRef) o, jdoClass.jormclass);
                id = JormPathHelper.getPath(((ClassRef) o));
                if (nd.isFieldName()) {
                    jormconfig.put(id, PNAME_SINGLE_BINDER_CLASS_NAME
                        + getCodingType(nd.getFieldName(), jdoClass.jormclass));
                } else if (nd.isNameRef()) {
                    jormconfig.put(id, getBinderClassName(nd, false));
                }
            } else if (o instanceof GenClassRef) {
                GenClassRef gcr = (GenClassRef) o;
                nd = getRefNameDef(gcr, jdoClass.jormclass);
                id = JormPathHelper.getPath((GenClassRef) o, false);
                if (nd.isFieldName()) {
                    jormconfig.put(id, PNAME_SINGLE_BINDER_CLASS_NAME
                        + getCodingType(nd.getFieldName(), jdoClass.jormclass));
                } else if (nd.isNameRef()) {
                    jormconfig.put(id, getBinderClassName(nd, true));
                }
                while (gcr.isGenClassRef()) {
                    gcr = gcr.getGenClassRef();
                    nd = getElemNameDef(gcr, jdoClass.jormclass);
                    id = JormPathHelper.getPath(gcr, false);
                    if (nd.isFieldName()) {
                        jormconfig.put(id, PNAME_SINGLE_BINDER_CLASS_NAME
                            + getCodingType(nd.getFieldName(), gcr));
                    } else if (nd.isNameRef()) {
                        jormconfig.put(id, getBinderClassName(nd, true));
                    }
                }
                if (gcr.isClassRef()) {
                    ClassRef cr = gcr.getClassRef();
                    nd = getElemNameDef(gcr, jdoClass.jormclass);
                    id = JormPathHelper.getPath(cr);
                    if (nd.isFieldName()) {
                        jormconfig.put(id, PNAME_SINGLE_BINDER_CLASS_NAME
                            + getCodingType(nd.getFieldName(), gcr));
                    } else if (nd.isNameRef()) {
                        jormconfig.put(id, getBinderClassName(nd, false));
                    }
                }
            }
        }

        //Build the array of String with the calculated Map
        StringBuffer jc = new StringBuffer("new String[][]");
        boolean first = true;
        for (Iterator it = jormconfig.entrySet().iterator(); it.hasNext();) {
            Map.Entry me = (Map.Entry) it.next();
            if (first) {
                jc.append('{');
                first = false;
            } else {
                jc.append(", ");
            }
            jc.append("{\"");
            jc.append(me.getKey());
            jc.append("\", \"");
            jc.append(me.getValue());
            jc.append("\"}");
        }
        jc.append('}');
        return jc.toString();
    }

    private String getBinderClassName(NameDef nd, boolean isGenClass) {
        CompositeName cn = nd.getNameRef().getCompositeName();
        String res = NamingRules.binderName(cn.getFQName());
        if (cn.getFQName().equals(SPEEDO_ID_NAME) && !isGenClass) {
            res += "Info";
        }
        return res;
    }

    private short getCodingType(String fn, MetaObject mo) {
        TypedElement te = null;
        if (mo instanceof Class) {
            te = ((Class) mo).getTypedElement(fn);
        } else if (mo instanceof GenClassRef) {
            te = ((GenClassRef) mo).getHiddenField(fn);
        } else {
            return -1;
        }
        return CTHelper.ptc2ct(te.getType().getTypeCode());
    }

    public class Field {
        public String name;
        public long number;
        public String getter;
        public String setter;
        public String modifier;
        public boolean isClassical;
        public boolean isContainerId;
        public boolean isArray;
        public String declaration;
        public boolean isKey;
        public String jormMeth;
        public int jormFieldIdLongPos;
        public long jormFieldId;
        public String jormFielIdDecl;
        public String jormType;
        public String  linkedField;
        public String memoryType;
        public boolean isMultiValued = false;
        public boolean isReference = false;
        public String toMemory = null;
        public String toStorage = null;
        public String userFieldMapping = null;

        /**
         * The name of the SpeedoGenClassProxy class to use
         */
        public String gcproxy;

        public boolean getIsReference() {
            return isReference;
        }

        public int getJormFieldIdLongPos() {
            return jormFieldIdLongPos;
        }

        public String getGcproxy() {
            return gcproxy;
        }

        public String getName() {
            return name;
        }

        public long getNumber() {
            return number;
        }

        public String getGetter() {
            return getter;
        }

        public String getSetter() {
            return setter;
        }

        public String getModifier() {
            return modifier;
        }

        public boolean isClassical() {
            return isClassical;
        }

        public boolean isContainerId() {
            return isContainerId;
        }

        public boolean isArray() {
            return isArray;
        }

        public String getDeclaration() {
            return declaration;
        }

        public boolean isKey() {
            return isKey;
        }

        public String getJormMeth() {
            return jormMeth;
        }

        public long getJormFieldId() {
            return jormFieldId;
        }

        public String getJormType() {
            return jormType;
        }

        public String getLinkedField() {
            return linkedField;
        }

        public String getMemoryType() {
            return memoryType;
        }

        public boolean getMultiValued() {
            return isMultiValued;
        }

        public String getJormFielIdDecl() {
            return jormFielIdDecl;
        }

        public String getToMemory() {
            return toMemory;
        }

        public String getToStorage() {
            return toStorage;
        }

        public String getUserFieldMapping() {
            return userFieldMapping;
        }
    }

    private String getValueGetter(java.lang.Class c) {
        return c.getName() + "Value";
    }

    public final java.lang.Class[][] PRIMITIVE_TYPES = {
        {Boolean.TYPE, Boolean.class},
        {Byte.TYPE, Byte.class},
        {Character.TYPE, Character.class},
        {Short.TYPE, Short.class},
        {Integer.TYPE, Integer.class},
        {Long.TYPE, Long.class},
        {Float.TYPE, Float.class},
        {Double.TYPE, Double.class}
    };

    private String getJavaLangType(java.lang.Class c) {
        for(int i=0; i<PRIMITIVE_TYPES.length; i++) {
            if (c == PRIMITIVE_TYPES[i][0]) {
                return PRIMITIVE_TYPES[i][1].getName();
            }
        }
        return null;
    }
}
