/**
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.generation.jorm;

import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.metadata.SpeedoField;
import org.objectweb.speedo.metadata.SpeedoIdentity;
import org.objectweb.speedo.metadata.SpeedoCollection;
import org.objectweb.speedo.metadata.SpeedoMap;
import org.objectweb.speedo.metadata.SpeedoExtension;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.api.UserFieldMapping;
import org.objectweb.speedo.generation.lib.NamingRules;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.ClassProject;
import org.objectweb.jorm.metainfo.api.Mapping;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.NameRef;
import org.objectweb.jorm.metainfo.api.CompositeName;
import org.objectweb.jorm.metainfo.api.ClassMapping;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.GenClassMapping;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.ScalarField;
import org.objectweb.jorm.metainfo.api.HelperClassMapping;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.lib.MetaInfoPrinter;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.asm.Type;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Collection;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.List;

/**
 * This class is a builder of jorm meta information. Its entries are the folowing:
 * <UL>
 * <LI>the jorm meta information manager,</LI>
 * <LI>A collection of SpeedoClass object,</LI>
 * <LI>the project name,</LI>
 * <LI>the mapper name,</LI>
 * <LI>a JormMIMappingBuilder instance able to build the mapping part for the
 * mapper which the name is specified.</LI>
 * </UL>
 * @author S.Chassande-Barrioz
 */
public class JormMIBuilder {

    /**
     * is the prefix of the fields used to identify a generic class
     */
    public final static String GENCLASS_ID_PREFIX = "id_";

    /**
     * is the prefix of the fields used for the element of a generic class
     */
    public final static String GENCLASS_ELM_PREFIX = "elem_";

    /**
     * is the name of the composite name used for the identifiers in case of
     * container identifier management.
     */
    public final static String SPEEDO_ID_NAME
            = "org.objectweb.jorm.facility.naming.polymorphid.PolymorphId";

    public final static String DEFAULT_RDB_BUILDER
            = "org.objectweb.speedo.generation.jorm.OneTableRdbJormIMappingBuilder";

    /**
     * is a name of field of the composite name used for the identifiers in case of
     * container identifier management.
     */
    public final static String SPEEDO_ID_OID = "objectId";
    /**
     * is a name of field of the composite name used for the identifiers in case of
     * container identifier management.
     */
    public final static String SPEEDO_ID_CID = "classId";

    /**
     * is the jorm meta information manager iinside which the Class and the
     * composite name must be defined.
     */
    private Manager manager = null;

    private Logger logger = null;

    /**
     * Builds a JormMIBuilder without a jorm meta information manager
     * and a logger.
     */
    public JormMIBuilder() {
    }

    /**
     * Builds a JormMIBuilder with a jorm meta information manager and a logger
     * @param manager
     */
    public JormMIBuilder(Manager manager, Logger logger) {
        this.manager = manager;
        this.logger = logger;
    }

    /**
     * retrieves the jorm meta information manager hosting the Class and
     * CompositeName instances.
     */
    public Manager getManager() {
        return manager;
    }

    /**
     * retrieves the jorm meta information manager hosting the Class and
     * CompositeName instances.
     */
    public void setManager(Manager manager) {
        this.manager = manager;
    }

    public Logger getLogger() {
        return logger;
    }

    public void setLogger(Logger logger) {
        this.logger = logger;
    }

    /**
     * Creates the jorm meta information for a set of persistent classes. Only
     * the generic part will be created.
     *
     * @param scs is a list of SpeedoClass instances.
     * @return a Collection of jorm meta object composed by Class instances and
     * CompositeName instances.
     */
    public Collection createMI(List scs) throws SpeedoException, PException {
        return createMI(scs, null, null);
    }

    /**
     * Creates the jorm meta information for a set of persistent classes.
     * @param scs is a list of SpeedoClass instances.
     * @param projectName is the project name for which the mapping must
     * be defined. If the value is null no mapping will be generated.
     * @param mapperName is the mapper name for which the mapping must
     * be defined. If the value is null no mapping will be generated.
     * @return a Collection of jorm meta object composed by Class instances and
     * CompositeName instances.
     */
    public Collection createMI(List scs,
                               String projectName,
                               String mapperName)
            throws SpeedoException, PException {
        JormMIMappingBuilder jmimb = null;
        String buildercn;
        if (mapperName != null) {
            if (mapperName.startsWith("rdb")) {
                buildercn = DEFAULT_RDB_BUILDER;
            } else {
                throw new SpeedoException(
                        "No default JormMIMappingBuilder for the mapper "
                        + mapperName);
            }
            try {
                jmimb = (JormMIMappingBuilder)
                        java.lang.Class.forName(buildercn).newInstance();
            } catch (Exception e) {
                throw new SpeedoException("Impossible to instanciate a the class "
                        + buildercn);
            }
            if (jmimb instanceof Loggable) {
                ((Loggable) jmimb).setLogger(logger);
            }
        }
        return createMI(scs, projectName, mapperName, jmimb);
    }
    /**
     * Creates the jorm meta information for a set of persistent classes.
     * @param scs is a list of SpeedoClass instances.
     * @param projectName is the project name for which the mapping must
     * be defined
     * @param mapperName is the mapper name for which the mapping must
     * be defined
     * @param mb is the build of the mapping part of the meta information
     * @return a Collection of jorm meta object composed by Class instances and
     * CompositeName instances.
     */
    public Collection createMI(List scs,
                               String projectName,
                               String mapperName,
                               JormMIMappingBuilder mb)
            throws SpeedoException, PException {
        if (manager == null) {
            throw new SpeedoException(
                    "the Jorm meta information manager must be assigned");
        }
        ArrayList createdMOs = new ArrayList(scs.size() * 2);
        SpeedoClass sc;
        String mn;
        if (mapperName != null) {
            // take the mapper name without the sub mapper name
            int idx = mapperName.indexOf(".");
            mn = (idx != -1 ? mapperName.substring(0, idx) : mapperName);
        } else {
            mn = null;
        }
        for(int i=0; i<scs.size(); i++) {
            sc = (SpeedoClass) scs.get(i);
            createJormClass(sc, projectName, mn, mb, createdMOs);
        }
        for(int i=0; i<scs.size(); i++) {
            sc = (SpeedoClass) scs.get(i);
            createReferences(sc, projectName, mn, mb, createdMOs);
        }
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            MetaInfoPrinter mip = new MetaInfoPrinter(manager);
            for(int i=0; i<createdMOs.size(); i++) {
                mip.print("", (MetaObject) createdMOs.get(i), logger);
            }
        }
        return createdMOs;
    }

    /**
     * Creates the jorm meta information for a persistent class. Only the
     * primitive fields are defined in the meta information. To define the
     * references of a persistent object the 'createReferences' method must be
     * used.
     * @param sc is the Speedo meta object describing the class which the jorm
     * meta information must be built.
     * @param projectName is the project name for which the mapping must
     * be defined
     * @param mapperName is the mapper name for which the mapping must
     * be defined
     * @param mb is the build of the mapping part of the meta information
     * @param createdMOs is a result paramter. This collection must be fill with
     * the created Jorm Meta objects representing  a class or a composite name.
     * Here the meta object of the class will be added. In addition if the name
     * def of the class is based on a composite name then its meta object will
     * be added too.
     */
    private void createJormClass(SpeedoClass sc,
                            String projectName,
                            String mapperName,
                            JormMIMappingBuilder mb,
                            Collection createdMOs)
            throws SpeedoException, PException {
        Class clazz = manager.getClass(sc.getFQName());
        if (clazz == null) {
            clazz = manager.createClass(sc.getFQName());
            createdMOs.add(clazz);
        }
        sc.jormclass = clazz;
        ClassMapping cm = null;
        if (projectName != null && mapperName != null) {
            ClassProject cp = clazz.getClassProject(projectName);
            if (cp == null) {
                cp = clazz.createClassProject(projectName);
            }
            Mapping mapping = cp.getMapping(mapperName);
            if (mapping != null) {
                //If the mapping is already defined then that means the jorm meta
                // information has been already defined.
                // => no more meta object has been added.
                return;
            }
            mapping = cp.createMapping(mapperName);
            cm = mb.createClassMapping(clazz, sc, mapping);
        }
        logger.log(BasicLevel.DEBUG, "Generate the Jorm MI for the class "
                + clazz.getFQName());

        Iterator fieldsIt = sc.jdoField.values().iterator();
        while (fieldsIt.hasNext()) {
            SpeedoField sp = (SpeedoField) fieldsIt.next();
            SpeedoExtension se = sp.getExtensionByKey(
                    SpeedoProperties.FIELD_CONVERTER);
            PType ptype = null;
            if (se != null) {
                try {
                    UserFieldMapping ufm = (UserFieldMapping)
                            java.lang.Class.forName(se.value).newInstance();
                    ptype = getPrimitivePType(Type.getType(ufm.getStorageType()));
                } catch (Exception e) {
                    throw new SpeedoException(
                            "Impossible to instanciate the UserFieldMapping class '"
                            + se.value + "' for the field '" + sp.name
                            + "' of the class '" + sc.getFQName() + "':", e);
                }
            } else {
                ptype = getPrimitivePType(Type.getType(sp.desc));
            }
            if (ptype != null) {
                logger.log(BasicLevel.DEBUG, "primitive field: " + sp.name + " / javatype: " + ptype.getJavaName());
                PrimitiveElement pe = clazz.createPrimitiveElement(
                        sp.name, ptype, PType.NOSIZE);
                if (cm != null && mb != null) {
                    mb.createFieldMapping(pe, sp, cm);
                }
            }
        }

        NameDef classNd = clazz.createNameDef();
        fillNameDef(classNd, sc, sc, clazz, null, cm, mb, true, false, sc.identityType == SpeedoIdentity.CONTAINER_ID, createdMOs);
        if (cm != null) {
            cm.createIdentifierMapping(classNd);
        }
    }

    /**
     * Builds the jorm meta objects representing the references of a persitent
     * class. It is supposed that if the reference targets a persistent class,
     * then the Class meta object and its NameDef are already defined.
     * @param sc is the Speedo meta object describing the class which the jorm
     * meta information must be built.
     * @param projectName is the project name for which the mapping must
     * be defined
     * @param mapperName is the mapper name for which the mapping must
     * be defined
     * @param mb is the build of the mapping part of the meta information
     * @param createdMOs is a result paramter. This collection must be fill with
     * the created Jorm Meta objects representing  a class or a composite name.
     * Here only the new used composite name will be added.
     */
    private void createReferences(SpeedoClass sc,
                                 String projectName,
                                 String mapperName,
                                 JormMIMappingBuilder mb,
                                 Collection createdMOs)
            throws SpeedoException, PException {
        Class clazz = manager.getClass(sc.getFQName());
        logger.log(BasicLevel.DEBUG, "Generate the Jorm MI for the references of class "
                + clazz.getFQName());
        Mapping mapping = null;
        ClassMapping cm = null;
        if (projectName != null && mapperName != null) {
            mapping = clazz.getClassProject(projectName)
                .getMapping(mapperName);
            cm = mapping.getClassMapping();
        }
        Iterator fieldsIt = sc.jdoField.values().iterator();
        //System.out.println("Manage reference of the class " + sc.getFQName());
        while (fieldsIt.hasNext()) {
            SpeedoField sp = (SpeedoField) fieldsIt.next();
            Type t = Type.getType(sp.desc);
            if (getPrimitivePType(t) != null) {
                //the primitive element is already managed before(createJormClass)
                continue;
            }
            String javatype = t.getClassName();
            if (!isGenClassRef(javatype)) {
                //                     reference to a class                   //
                //============================================================//
                logger.log(BasicLevel.DEBUG, "Class reference field: " + sp.name + " / javatype: " + javatype);
                // reference to a class
                SpeedoClass tsc = sc.jdoPackage.jdoXMLDescriptor
                        .getSpeedoClass(javatype);
                if (tsc == null) {
                    throw new SpeedoException("No persistent class '"
                            + javatype + "' found in the file: "
                            + sc.jdoPackage.jdoXMLDescriptor.xmlFile);
                }
                Class tclass = manager.getClass(tsc.getFQName());
                if (tclass == null) {
                    manager.createClass(tsc.getFQName());
                }
                ClassRef cr = clazz.createClassRef(sp.name, tclass);
                NameDef refNd = cr.createRefNameDef();
                fillNameDef(refNd, tsc, sc, clazz, cr, cm, mb, false, false, true, createdMOs);
                cm.createReferenceMapping(sp.name, refNd);

            } else {
                //                reference to a generic class                //
                //============================================================//
                String innerType = getInnerType(sp);
                logger.log(BasicLevel.DEBUG, "Multivalued field: " + sp.name + " / javatype: " + javatype + " / inner element: " + innerType);
                GenClassRef gcr = clazz.createGenClassRef(sp.name, javatype);
                GenClassMapping gcm = null;
                if (mapping != null) {
                    gcm = mb.createGenClassMapping(gcr, sp, mapping);
                }

                // reference to a generic class from the class
                NameDef refNd = gcr.createRefNameDef();
                fillNameDef(refNd, sc, sc, clazz, gcr, cm, mb, false, false, false, createdMOs);
                if (cm != null) {
                    cm.createReferenceMapping(sp.name, refNd);
                }

                //identifier of the gen class
                NameDef gcidNd = gcr.createIdNameDef();
                fillNameDef(gcidNd, sc, sc, gcr, gcr, gcm, mb, true, true, true, createdMOs);
                if (gcm != null) {
                    gcm.createIdentifierMapping(gcidNd);
                }

                //element of the gen class
                PType type = getPrimitivePType(innerType);
                if (type != null) {
                    // gen class of primitive type
                    PrimitiveElement pe = gcr.createPrimitiveElement(
                            type, PType.NOSIZE);
                    if (gcm != null) {
                        mb.createFieldMapping(pe, null, gcm);
                    }

                } else if (!isGenClassRef(innerType)) {
                    // gen class of classref
                    SpeedoClass tsc = sc.jdoPackage.jdoXMLDescriptor
                            .getSpeedoClass(innerType);

                    if (tsc == null) {
                        tsc = (SpeedoClass)
                                sc.jdoPackage.jdoClass.get(innerType);
                    }
                    if (tsc == null) {
                        throw new SpeedoException("The persistent class '"
                                + sc.getFQName()
                                + "' tries to reference (througth the field '"
                                + sp.name
                                + "') the class '" + innerType
                                + "' not defined in the .jdo file '"
                                + sc.jdoPackage.jdoXMLDescriptor.xmlFile + "'");
                    }
                    Class tclass = manager.getClass(tsc.getFQName());
                    if (tclass == null) {
                        throw new SpeedoException("No Class '"
                                + tsc.getFQName()
                                + "' defined in the jorm meta information, found:"
                                + manager.getClasses());
                    }
                    ClassRef cr = gcr.createClassRef(tclass);
                    NameDef elemNd = cr.createRefNameDef();
                    fillNameDef(elemNd, tsc, sc, gcr, gcr, gcm, mb, false, true, true, createdMOs);
                    if (gcm != null) {
                        gcm.createReferenceMapping(sp.name, elemNd);
                    }
                } else {
                    throw new SpeedoException(
                            "unmanaged generic class element: " + innerType);
                }
            }
        }

    }

    /**
     * Fill the name def of an identifier or a reference
     * @param nd is the name def to fill
     * @param isIdentifier indicates if the name represents an identifier (true)
     * or a reference (false).
     * @param isInGenClass indicates if the name is defined in a generic class
     * (true) of in a class (false).
     * @param ref meta object which the name def must be
     * defined. This value is used only in the case of isIdentifier == false and
     * isInGenClass == false.
     * @param tsc is the speedo meta object representing the referenced class.
     * This value is used only in the case of isIdentifier == false.
     * @param mo is the jorm meta object hosting the name def and on which the
     * eventual hidden field will be created.
     * @param hcm if the mapping structure hosting the mapping of the reference.
     * This value must be ClassMapping or GenClassMapping instance.
     * @param mb is the mapping builder permitting the creation of the mapping
     * part.
     * @param createdMOs is a result paramter. This collection must be fill with
     * the created Jorm Meta objects representing  a class or a composite name.
     * Here only the new used composite name will be added.
     * @throws SpeedoException if the speedo meta information is not completly
     * defined.
     */
    private void fillNameDef(NameDef nd,
                             SpeedoClass tsc,
                             SpeedoClass ssc,
                             MetaObject mo,
                             Reference ref,
                             HelperClassMapping hcm,
                             JormMIMappingBuilder mb,
                             boolean isIdentifier,
                             boolean isInGenClass,
                             boolean createField,
                             Collection createdMOs)
            throws SpeedoException, PException {
        if (tsc.identityType == SpeedoIdentity.USER_ID) {
            ArrayList idFields = new ArrayList();
            Iterator fieldsIt = tsc.jdoField.values().iterator();
            while (fieldsIt.hasNext()) {
                SpeedoField sp = (SpeedoField) fieldsIt.next();
                if (sp.primaryKey) {
                    idFields.add(sp);
                }
            }
            String objectidClass = tsc.objectidClass;
            if (tsc.generateObjectId()) {
                objectidClass = NamingRules.generatedObjectIdName(tsc.getFQName());
            }
            if (objectidClass == null || objectidClass.length() == 0) {
                //No object id specified
                if (idFields.size() == 0) {
                    throw new SpeedoException(
                            getErrorMessage(tsc, mo, ref)
                            + "no primary key field and no object id defined");
                }
                if (idFields.size() > 1) {
                    String fn = "[";
                    String sep = "";
                    for(int i=0; i<idFields.size(); i++) {
                        fn += sep + ((SpeedoField) idFields.get(i)).name;
                        sep = ",";
                    }
                    fn += "]";
                    throw new SpeedoException(
                            getErrorMessage(tsc, mo, ref)
                            + "there are several primary key fields "
                            + fn
                            + " and no an object id class.");
                }
                SpeedoField sp = (SpeedoField) idFields.get(0);
                String clafn = getNameDefFieldPrefix(
                        ref, isIdentifier, isInGenClass) + sp.name;
                if (createField) {
                    createNameDefField(mo, clafn,
                            getPrimitivePType(Type.getType(sp.desc)));
                }
                nd.setFieldName(clafn);
            } else {
                //The user has specified an object id ==> use it for the namedef
                CompositeName cn = manager.getCompositeName(objectidClass);
                boolean cnnotdefined = (cn == null);
                if (cnnotdefined) {
                    cn = manager.createCompositeName(objectidClass);
                    createdMOs.add(cn);
                }
                NameRef nr = nd.createNameRef(cn);
                for (int i = 0; i < idFields.size(); i++) {
                    SpeedoField sf = (SpeedoField) idFields.get(i);
                    if (cnnotdefined) {
                        cn.createCompositeNameField(sf.name,
                                getPrimitivePType(Type.getType(sf.desc)),
                                PType.NOSIZE);
                    }
                    String fn = getNameDefFieldPrefix(
                            ref, isIdentifier, isInGenClass) + sf.name;
                    if (createField) {
                        createNameDefField(mo, fn,
                                getPrimitivePType(Type.getType(sf.desc)));
                    }
                    nr.addProjection(sf.name, fn);
                }
            }
        } else {
            String prefix = getNameDefFieldPrefix(ref, isIdentifier, isInGenClass);
            PrimitiveElement oid;
            String oidname = prefix + "o_id";
            String cidname = prefix + "c_id";
            PrimitiveElement cid;
            if (createField) {
                oid = createNameDefField(mo, oidname, PTypeSpace.LONG);
                cid = createNameDefField(mo, cidname, PTypeSpace.LONG);
            } else {
                cid = getPrimitiveField(mo, cidname);
                if (cid == null) {
                    throw new SpeedoException(
                            getErrorMessage(tsc, mo, ref)
                            + " Impossible to get the field '" + cidname + "'");
                }
                oid = getPrimitiveField(mo, oidname);
                if (cid == null) {
                    throw new SpeedoException(
                            getErrorMessage(tsc, mo, ref)
                            + " Impossible to get the field '" + oidname + "'");
                }
            }
            CompositeName speedoidcn =  getPolymorphIdCN();
            NameRef nr = nd.createNameRef(speedoidcn);
            nr.addProjection(SPEEDO_ID_OID, oid.getName());
            nr.addProjection(SPEEDO_ID_CID, cid.getName());
        }
        if (createField) {
            mb.createNameDefMapping(hcm, nd, ssc, isIdentifier, isInGenClass);
        }

    }

    /**
     * Calculates the prefix of a field use in a name def.
     * @param ref is the meta object of the reference if the namedef is
     * used for  a reference.
     * @param isIdentifier indicates if the namedef is used for an identifier
     * (true) or if the namedef is used for a reference (false).
     * @param isInGenClass indicates if the namedef is used in a Generic class
     * (true) or if the namedef is used in a class (false).
     * @return a string value (never null) representing the prefix of a name
     * def field.
     */
    private String getNameDefFieldPrefix(Reference ref,
                                         boolean isIdentifier,
                                         boolean isInGenClass) {
        if (isIdentifier) {
            return "";
        } else {
            if (isInGenClass) {
                return GENCLASS_ELM_PREFIX;
            } else {
                if (ref instanceof GenClassRef) {
                    //The reference of the GCR is mapped over the class identifer
                    // fields
                    return "";
                } else {
                    return ref.getName() + '_';
                }
            }
        }
    }

    /**
     * It creates a field of a name def. This field will be hidden
     * @param mo is the jorm meta object hosting the field to create
     * @param fn is the name of the field to create
     * @param type is the type of the field to create
     * @return the jorm meta object representing the namedef field
     */
    private ScalarField createNameDefField(MetaObject mo,
                                          String fn,
                                          PType type)
            throws SpeedoException {
        ScalarField sf;
        if (mo instanceof Class) {
            sf = ((Class) mo).createHiddenField(fn, type, PType.NOSIZE);
        } else if (mo instanceof GenClassRef) {
            sf = ((GenClassRef) mo).createHiddenField(fn, type, PType.NOSIZE);
        } else {
            throw new SpeedoException(
                    "Impossible to create hidden field on this meta object: " + mo);
        }
        return sf;
    }

    /**
     * Retrieves the jorm type matching to primitive type. It converts a Type
     * defined in ASM into a PType defined in Jorm. If the type is not a jorm
     * primitive type then a null value is returned.
     */
    private PType getPrimitivePType(Type t) {
        switch (t.getSort()) {
        case Type.BOOLEAN:
            return PTypeSpace.BOOLEAN;
        case Type.CHAR:
            return PTypeSpace.CHAR;
        case Type.BYTE:
            return PTypeSpace.BYTE;
        case Type.SHORT:
            return PTypeSpace.SHORT;
        case Type.INT:
            return PTypeSpace.INT;
        case Type.FLOAT:
            return PTypeSpace.FLOAT;
        case Type.LONG:
            return PTypeSpace.LONG;
        case Type.DOUBLE:
            return PTypeSpace.DOUBLE;
        case Type.OBJECT:
            String cn = t.getClassName();
            if (!isGenClassRef(cn)) {
                return getPrimitivePType(cn);
            }
        case Type.ARRAY:
        default:
            //System.out.println("getPrimitivePType(" + t.getClassName() + "): return null");
            return null;
        }
    }

    public static PType getPrimitivePType(String cn) {
        for (int i = 0; i < PTypeSpace.PREDEFINEDPTYPES.length; i++) {
            if (PTypeSpace.PREDEFINEDPTYPES[i].getJavaName().equals(cn)) {
                return PTypeSpace.PREDEFINEDPTYPES[i];
            }
        }
        return null;
    }

    /**
     * Indicates if the string representing a java type is a generic class (
     * java.util.Collection | java.util.Set | java.util.Map
     * @param javatype
     * @return
     */
    private boolean isGenClassRef(String javatype) {
        return Collection.class.getName().equals(javatype)
                || Set.class.getName().equals(javatype)
                || Map.class.getName().equals(javatype)
                ;
    }

    /**
     * Retrieves the type of a generic class element.
     * @param sp is the field which represents the generic class
     * @return a String value representing the java type of the element, or null
     * if the Speedofield does not represents a generic class.
     */
    private String getInnerType(SpeedoField sp) {
        if (sp.jdoTuple instanceof SpeedoCollection) {
            return (String) ((SpeedoCollection) sp.jdoTuple).elementType;
        } else if (sp.jdoTuple instanceof SpeedoMap) {
            return (String) ((SpeedoMap) sp.jdoTuple).valueType;
        }
        return null;
    }


    /**
     * retrieves the start of a pretty error message.
     */
    private String getErrorMessage(SpeedoClass sc, MetaObject mo, Reference ref) {
        String fqcn;
        if (mo == null) {
            fqcn = "null";
        } else if (mo instanceof Class) {
            fqcn = ((Class) mo).getFQName();
        } else if (mo instanceof GenClassRef) {
            fqcn = ((GenClassRef) mo).getGenClassId();
        } else {
            fqcn = mo.toString();
        }
        String res = "Impossible to define an user identifier for the ";
        if (ref == null) {
            res += "identifier of the class '" + fqcn;
        } else {
            res += "reference '" + ref.getName()
                    + "' from the class '" + fqcn
                    + "' to the class '" + sc.getFQName();
        }
        return res  + "': ";
    }

    private CompositeName getPolymorphIdCN() {
        CompositeName cn = manager.getCompositeName(SPEEDO_ID_NAME);
        if (cn == null) {
            cn = manager.createCompositeName(SPEEDO_ID_NAME);
            cn.createCompositeNameField(SPEEDO_ID_CID, PTypeSpace.LONG, PType.NOSIZE);
            cn.createCompositeNameField(SPEEDO_ID_OID, PTypeSpace.LONG, PType.NOSIZE);
        }
        return cn;
    }

    private PrimitiveElement getPrimitiveField(MetaObject mo, String name) {
        if (mo instanceof Class) {
            return (PrimitiveElement) ((Class) mo).getTypedElement(name);
        } else if (mo instanceof GenClassRef) {
            return ((GenClassRef) mo).getHiddenField(name);
        } else {
            return null;
        }
    }
}
