/**
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.generation.jorm;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassMapping;
import org.objectweb.jorm.metainfo.api.GenClassMapping;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.HelperClassMapping;
import org.objectweb.jorm.metainfo.api.Mapping;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.PrimitiveElementMapping;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.rdb.api.RdbClassMultiMapping;
import org.objectweb.jorm.metainfo.rdb.api.RdbGenClassMapping;
import org.objectweb.jorm.metainfo.rdb.api.RdbTable;
import org.objectweb.jorm.metainfo.rdb.api.RdbExternalTable;
import org.objectweb.jorm.metainfo.rdb.api.RdbJoin;
import org.objectweb.jorm.metainfo.rdb.api.RdbPrimitiveElementMapping;
import org.objectweb.jorm.metainfo.rdb.lib.BasicRdbClassMultiMapping;
import org.objectweb.jorm.metainfo.rdb.lib.BasicRdbGenClassMapping;
import org.objectweb.jorm.runtime.JormPathHelper;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.metadata.SpeedoExtension;
import org.objectweb.speedo.metadata.SpeedoField;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;
import java.util.StringTokenizer;

/**
 * This class is an implementation of the JormMIMappingBuilder for the mapper
 * rdb and its sub mappers. It defines a simple mapping where each persistent
 * objects (generic class included) are mapped over one table.
 *
 * @author S.Chassande-Barrioz
 */
public class OneTableRdbJormIMappingBuilder
        implements JormMIMappingBuilder, SpeedoProperties, Loggable {

    private final static String CLASS_MAPPING_RULE_NAME = "multi-table";
    private final static String GENCLASS_MAPPING_RULE_NAME = "to-table";

    private Logger logger;

    // IMPLEMENTATION OF THE Loggable INTERFACE //
    //------------------------------------------//

    public Logger getLogger() {
        return logger;
    }

    public LoggerFactory getLoggerFactory() {
        return null;
    }

    public void setLogger(Logger logger) {
        this.logger = logger;
    }

    public void setLoggerFactory(LoggerFactory loggerFactory) {
        if (logger == null && loggerFactory != null) {
            logger = loggerFactory.getLogger(
                    "org.objectweb.speedo.generation.jorm.mimappingbuilder");
        }
    }

    // IMPLEMENTATION OF THE Loggable INTERFACE //
    //------------------------------------------//

    /**
     * It builds a RdbClassMultiMapping and creates the main table.
     * The name of the table is built with the short class name.
     * @param clazz is the Jorm meta object representing the class which the
     * ClassMapping must be built.
     * @param mapping is the Mapping instance which will host the ClassMapping.
     * @return the ClassMapping instance built by the method (never null).
     * @throws PException if it is not possible to build the ClassMapping
     */
    public ClassMapping createClassMapping(Class clazz,
                                           SpeedoClass sc,
                                           Mapping mapping)
            throws PException {
        RdbClassMultiMapping rcm = new BasicRdbClassMultiMapping(
                CLASS_MAPPING_RULE_NAME, clazz, mapping);
        mapping.setClassMapping(rcm);
        SpeedoExtension se =
            (sc == null ? null : sc.getExtensionByKey(SQL_NAME));
        rcm.createRdbTable((se == null ? getTableName(sc) : se.value));
        return rcm;
    }

    /**
     * It builds a RdbClassMultiMapping and creates the table.
     * the table is build with the generic class identifier ("A/f1" ==> "A_f1")
     * @param gcr is the Jorm meta object representing the gen class which the
     * GenClassMapping must be built.
     * @param mapping is the Mapping instance which will host the GenClassMapping.
     * @return the GenClassMapping instance built by the method (never null).
     * @throws PException if it is not possible to build the GenClassMapping
     */
    public GenClassMapping createGenClassMapping(GenClassRef gcr,
                                                 SpeedoField sf,
                                                 Mapping mapping)
            throws PException {
        RdbGenClassMapping rgcm = new BasicRdbGenClassMapping(
                GENCLASS_MAPPING_RULE_NAME, gcr, mapping);
        mapping.addGenClassMapping(gcr.getGenClassId(), rgcm);
        SpeedoExtension se = sf.getExtension(VENDOR_NAME, JOIN_TABLE);
        if (se != null) {
            rgcm.createRdbTable(se.value);
        } else if (gcr.isClassRef()
                && sf.getExtension(VENDOR_NAME, TARGET_FK) == null) {
            // The gen class is mapped over the table of the targeted class
            SpeedoClass sc = sf.jdoClass.jdoPackage.jdoXMLDescriptor
                    .getSpeedoClass(gcr.getClassRef().getMOClass().getFQName());
            se = (sc == null ? null : sc.getExtensionByKey(SQL_NAME));
            RdbTable t = rgcm.createRdbTable(
                    (se == null ? getTableName(sc) : se.value));
            if (sf.getExtensionByKey(REVERSE_FIELD) != null) {
                //optimisation: if there is a reverse field then we suppose that
                // the user will do the coherence.
                t.setReadOnly(true);
            }
            // The GC table and the table of the targeted class are colocated
            t.setColocated(true);
            Manager mgr = getManager(gcr);
            Class tclazz = mgr.getClass(sc.getFQName());
            RdbClassMultiMapping trcm = (RdbClassMultiMapping)
            tclazz.getClassProject(mapping.getClassMapping().getProjectName())
                .getMapping(mapping.getMapperName()).getClassMapping();
            trcm.getRdbTable().setColocated(true);
            trcm.getRdbTable().setColocatedMaster(true);
        } else {
            //default mapping: join table
            rgcm.createRdbTable(getTableName(gcr));
        }
        return rgcm;
    }

    /**
     * The column name is built with the field name.
     * @param pe is the Jorm meta object representing a primitive field
     * @param cm is the MappingStructure which will host the mapping of the
     * field
     * @return a PrimitiveElementMapping corresponding to the given primitive
     * field.
     * @throws PException if it is not possible to build the mapping of the
     * primitive field.
     */
    public PrimitiveElementMapping createFieldMapping(PrimitiveElement pe,
                                                      SpeedoField sf,
                                                      HelperClassMapping cm)
            throws PException {
        RdbTable t;
        if (cm instanceof RdbClassMultiMapping) {
            t = ((RdbClassMultiMapping) cm).getRdbTable();
        } else if (cm instanceof RdbGenClassMapping) {
            t = ((RdbGenClassMapping) cm).getRdbTable();
        } else {
            new Exception().printStackTrace();
            throw new PException("Unmanaged HelperClassMapping: " + cm);
        }
        SpeedoExtension se =
            (sf == null ? null : sf.getExtensionByKey(SQL_NAME));
        return t.createPrimitiveElementMapping(pe,
            (se == null ? getColumnName(pe) : se.value));
    }

    /**
     * It creates the mapping of a name def (reference or identifier).
     * @param cm is the MappingStructure which will host the mapping of the
     * fields of the name def
     * @param nd is the NameDef which must be mapped.
     * @throws PException if it is not possible to build the mapping of the
     * field.
     */
    public void createNameDefMapping(HelperClassMapping cm,
                                     NameDef nd,
                                     SpeedoClass sc,
                                     boolean isIdentifier,
                                     boolean isInGenClass)
            throws PException, SpeedoException {
        if (!isInGenClass) {
            Class clazz;
            if (isIdentifier) {
                //This is the identifier of the class
                clazz = (Class) nd.getParent();
                if (nd.isFieldName()) {
                    PrimitiveElement pe = (PrimitiveElement)
                            clazz.getTypedElement(nd.getFieldName());
                    if (pe == null) {
                        throw new SpeedoException("impossible to map the fields of the namedef of the metaobject '" + JormPathHelper.getPath(clazz) + "'.");
                    }
                    createFieldMapping(pe, null, cm);
                } else {
                    Iterator it = nd.getNameRef().getProjection().values().iterator();
                    while(it.hasNext()) {
                        PrimitiveElement pe = (PrimitiveElement)
                            clazz.getTypedElement((String) it.next());
                        if (pe == null) {
                            throw new SpeedoException("impossible to map the fields of the namedef of the metaobject '" + JormPathHelper.getPath(clazz) + "'.");
                        }
                        createFieldMapping(pe, null, cm);
                    }
                }
            } else {
                MetaObject mo = nd.getParent();
                if (mo instanceof GenClassRef) {
                    // The reference of the gen class is mapped over the source
                    // class identifier, then the field of the identifier are
                    // already mapped.
                    return;
                }
                //This is a reference to a persistent object since a class.
                createClassRefNameDef(sc, (ClassRef) mo, cm, nd);
            }
        } else {
            if (isIdentifier) {
                creaGenClassRefIdNameDef(nd, sc, (RdbGenClassMapping) cm);
            } else {
                creaGenClassRefRefNameDef(nd, sc, (RdbGenClassMapping) cm);
            }
        }
    }


    // PRIVATE METHODS //
    //-----------------//

    /**
     * Mappes a ClassRef over sql columns. In according to the speedo extension
     * a local foreign key (default case) or a join with the referenced table
     * is added for the reference mapping.
     * @param sc is the SpeedoClass hosting the reference
     * @param cr is the jorm meta object representing the reference
     * @param cm is the ClassMapping of the class hosting the reference
     * @param crnd is the name def of the reference.
     * @throws SpeedoException if the meta information is not complete
     */
    private void createClassRefNameDef(SpeedoClass sc,
                                       ClassRef cr,
                                       HelperClassMapping cm,
                                       NameDef crnd)
            throws SpeedoException, PException {
        logger.log(BasicLevel.DEBUG, "Create the name def of the ClassRef: "
                + JormPathHelper.getPath(cr));
        //The ClassMapping of the current class.
        RdbClassMultiMapping rcm = (RdbClassMultiMapping) cm;
        //The Speedo Meta object of the referenced class
        SpeedoClass tsc = sc.jdoPackage.jdoXMLDescriptor
            .getSpeedoClass(cr.getMOClass().getFQName());
        //The ClassMapping of the referenced class.
        RdbClassMultiMapping trcm = (RdbClassMultiMapping)
                tsc.jormclass.getClassProject(cm.getProjectName())
                .getMapping(cm.getMapperName()).getClassMapping();
        //The Speedo meta object mathcing to the ClassRef
        SpeedoField sf = (SpeedoField) sc.jdoField.get(cr.getName());
        if (sf == null) {
            throw new SpeedoException("impossible to map the namedef of the field '"
                    + JormPathHelper.getPath(cr) + "': no SpeedoField '"
                    + cr.getName()+ "' on the SpeedoClass " + sc.getFQName());
        }

        //The name def of the referenced class
        NameDef tclassnd = (NameDef)
                trcm.getIdentifierMapping().getLinkedMO();

        SpeedoExtension se = sf.getExtension(VENDOR_NAME, TARGET_FK);
        String sfk = (se == null ? null : se.value);
        if (sfk != null) {
            //foreign key in the external table

            se = tsc.getExtensionByKey(SQL_NAME);
            String tn = (se == null ? getTableName(sc) : se.value);
            logger.log(BasicLevel.DEBUG, "  - External table: " + tn);
            RdbExternalTable exttable = rcm.createRdbExternalTable(tn);
            if (sf.getExtensionByKey(REVERSE_FIELD) != null) {
                exttable.setReadOnly(true);
            }
            // The external table and the table of the targeted class are
            // colocated
            exttable.setColocated(true);
            trcm.getRdbTable().setColocated(true);
            trcm.getRdbTable().setColocatedMaster(true);

            //optimisation: if there is a reverse field then we suppose that
            // the user will do the coherence.
            se = sf.getExtension(VENDOR_NAME, REVERSE_FIELD);
            logger.log(BasicLevel.DEBUG, "  - readonly: " + (se != null));
            exttable.setReadOnly(se != null);

            //Define the join to the external table
            RdbJoin j = exttable.createRdbJoin(cr.getName());
            Iterator it = getPk2Fk(sfk).entrySet().iterator();
            while(it.hasNext()) {
                Map.Entry me = (Map.Entry) it.next();
                logger.log(BasicLevel.DEBUG, "  - join: "
                        + me.getKey() + " == " + me.getValue());
                j.addJoinColumnNames((String) me.getKey(),
                        (String) me.getValue());
            }

            //map the name def over pk field of the referenced class
            if (tclassnd.isFieldName()) {
                RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
                        trcm.getPrimitiveElementMapping(tclassnd.getFieldName());
                PrimitiveElement pe = (PrimitiveElement)
                        sc.jormclass.getTypedElement(crnd.getFieldName());
                exttable.createPrimitiveElementMapping(pe, pem.getName());
            } else {
                Map tclassndproj = tclassnd.getNameRef().getProjection();
                it = crnd.getNameRef().getProjection().entrySet().iterator();
                while(it.hasNext()) {
                    Map.Entry me = (Map.Entry) it.next();
                    PrimitiveElement pe = (PrimitiveElement)
                            sc.jormclass.getTypedElement((String) me.getValue());
                    RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
                        trcm.getPrimitiveElementMapping(
                                (String) tclassndproj.get(me.getKey()));
                    exttable.createPrimitiveElementMapping(pe, pem.getName());
                }
            }
        } else {
            //local foreign keys
            RdbTable table = rcm.getRdbTable();
            logger.log(BasicLevel.DEBUG, "  - main table: " + table.getName());

            String tfk = (se == null ? null : se.value);
            se = sf.getExtension(VENDOR_NAME, SOURCE_FK);
            Map pk2fk = null;
            if (tfk != null) {
                pk2fk = getPk2Fk(tfk);
            } else {
                //default mapping of a reference: local foreign keys
                // The column name is equal to the field name.
                logger.log(BasicLevel.DEBUG, "  - default mapping");
                pk2fk = new HashMap();
                if (crnd.isFieldName()) {
                    RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
                            trcm.getPrimitiveElementMapping(tclassnd.getFieldName());
                    pk2fk.put(pem.getName(), crnd.getFieldName());
                    logger.log(BasicLevel.DEBUG, "  - pk2fk: " + pem.getName() + " == " + crnd.getFieldName());
                } else {
                    Iterator it = crnd.getNameRef().getProjection().entrySet().iterator();
                    Map tclassndproj = tclassnd.getNameRef().getProjection();
                    while (it.hasNext()) {
                        Map.Entry me = (Map.Entry) it.next();
                        RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
                                trcm.getPrimitiveElementMapping((String)
                                tclassndproj.get(me.getKey()));
                        pk2fk.put(pem.getName(), me.getValue());
                    }
                }
            }
            //Map the name def of the reference over new local columns
            if (crnd.isFieldName()) {
                RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
                        trcm.getPrimitiveElementMapping(tclassnd.getFieldName());
                PrimitiveElement pe = (PrimitiveElement)
                        sc.jormclass.getTypedElement(crnd.getFieldName());
                table.createPrimitiveElementMapping(pe,
                        (String) pk2fk.get(pem.getName()));
            } else {
                Map tclassndproj = tclassnd.getNameRef().getProjection();
                Iterator it = crnd.getNameRef().getProjection().entrySet().iterator();
                while(it.hasNext()) {
                    Map.Entry me = (Map.Entry) it.next();
                    PrimitiveElement pe = (PrimitiveElement)
                            sc.jormclass.getTypedElement((String) me.getValue());
                    RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
                        trcm.getPrimitiveElementMapping(
                                (String) tclassndproj.get(me.getKey()));
                    table.createPrimitiveElementMapping(pe,
                            (String) pk2fk.get(pem.getName()));
                }
            }
        }
    }


    /**
     * Mappes the identifier of a GenClassRef over sql columns. In according
     * to the speedo extensions a join table is used (default case) or not.
     *
     * The columms match to the foreign key of the souce class. The default
     * column names of the join table are the name of the hidden field of the
     * genclass.
     *
     * @param sc is the SpeedoClass hosting the reference
     * @param nd is the name def of the gen class identifier
     * @param sc is the Speedo meta object hosting the field which references
     * the generic class.
     * @param rgcm is the jorm meta object representing the mapping of the gen
     * class.
     * @throws PException
     */
    private void creaGenClassRefIdNameDef(NameDef nd,
                                     SpeedoClass sc,
                                     RdbGenClassMapping rgcm)
            throws PException {
        MetaObject mo = nd.getParent();
        RdbTable table = rgcm.getRdbTable();
        GenClassRef gcr = (GenClassRef) mo;
        SpeedoField sf = (SpeedoField) sc.jdoField.get(gcr.getName());
        RdbClassMultiMapping srcm = (RdbClassMultiMapping)
                ((Class) gcr.getParent())
                .getClassProject(rgcm.getProjectName())
                .getMapping(rgcm.getMapperName()).getClassMapping();
        NameDef sclassnd = (NameDef) srcm.getIdentifierMapping().getLinkedMO();
        SpeedoExtension se = sf.getExtension(VENDOR_NAME, SOURCE_FK);
        Map pk2fk = (se == null
                ? null // default mapping
                : getPk2Fk(se.value));
        if (nd.isFieldName()) {
            PrimitiveElement pe = gcr.getHiddenField(nd.getFieldName());
            String colname;
            if (pk2fk == null) {
                // default mapping
                colname = pe.getName();
            } else {
                RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
                        srcm.getPrimitiveElementMapping(sclassnd.getFieldName());
                colname = (String) pk2fk.get(pem.getName());
            }
            table.createPrimitiveElementMapping(pe, colname);
        } else {
            Map sclassndproj = sclassnd.getNameRef().getProjection();
            Iterator it = nd.getNameRef().getProjection().entrySet().iterator();
            while (it.hasNext()) {
                Map.Entry me = (Map.Entry) it.next();
                PrimitiveElement pe =
                        gcr.getHiddenField((String) me.getValue());
                String colname;
                if (pk2fk == null) {
                    // default mapping
                    colname = pe.getName();
                } else {
                    RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
                            srcm.getPrimitiveElementMapping(
                                    (String) sclassndproj.get(me.getKey()));
                    colname = (String) pk2fk.get(pem.getName());
                }
                table.createPrimitiveElementMapping(pe, colname);
            }
        }
    }


    /**
     * Mappes the element of a GenClassRef over sql columns. In according
     * to the speedo extensions a join table is used (default case) or not.
     *
     * The element of the gen class is a Gen class (not supported) or
     * a classRef. In all case the parent of this element is the GenClassRef
     *
     * @param sc is the SpeedoClass hosting the reference
     * @param nd is the name def of the gen class element
     * @param sc is the Speedo meta object hosting the field which references
     * the generic class.
     * @param rgcm is the jorm meta object representing the mapping of the gen
     * class.
     * @throws PException
     */
    private void creaGenClassRefRefNameDef(NameDef nd,
                                     SpeedoClass sc,
                                     RdbGenClassMapping rgcm)
            throws PException {
        MetaObject mo = nd.getParent();
        RdbTable table = rgcm.getRdbTable();
        GenClassRef gcr = (GenClassRef) mo.getParent();
        ClassRef cr = (ClassRef) mo;
        SpeedoField sf = (SpeedoField) sc.jdoField.get(gcr.getName());
        RdbClassMultiMapping trcm = (RdbClassMultiMapping)
                cr.getMOClass().getClassProject(rgcm.getProjectName())
                .getMapping(rgcm.getMapperName()).getClassMapping();
        NameDef tclassnd = (NameDef) trcm.getIdentifierMapping().getLinkedMO();
        if (isColocatedGCTable(sf)) {
            //The columms match to the primary key of the target class
            if (nd.isFieldName()) {
                PrimitiveElement pe = gcr.getHiddenField(nd.getFieldName());
                RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
                        trcm.getPrimitiveElementMapping(tclassnd.getFieldName());
                //We must use the same column name
                table.createPrimitiveElementMapping(pe, pem.getName());
            } else {
                Map tclassndproj = tclassnd.getNameRef().getProjection();
                Iterator it = nd.getNameRef().getProjection().entrySet().iterator();
                while (it.hasNext()) {
                    Map.Entry me = (Map.Entry) it.next();
                    PrimitiveElement pe =
                            gcr.getHiddenField((String) me.getValue());
                    RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
                            trcm.getPrimitiveElementMapping(
                                    (String) tclassndproj.get(me.getKey()));
                    //We must use the same column name
                    table.createPrimitiveElementMapping(pe, pem.getName());
                }
            }
        } else {
            //The columns are foreign key of the primary key of the
            // target class.
            SpeedoExtension se = sf.getExtension(VENDOR_NAME, TARGET_FK);
            Map pk2fk = (se == null
                    ? null // default mapping
                    : getPk2Fk(se.value));
            if (nd.isFieldName()) {
                PrimitiveElement pe =
                        gcr.getHiddenField(nd.getFieldName());
                String colname;
                if (pk2fk == null) {
                    // default mapping
                    colname = pe.getName();
                } else {
                    RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
                            trcm.getPrimitiveElementMapping(tclassnd.getFieldName());
                    colname = (String) pk2fk.get(pem.getName());
                }
                table.createPrimitiveElementMapping(pe, colname);
            } else {
                Map tclassndproj = tclassnd.getNameRef().getProjection();
                Iterator it = nd.getNameRef().getProjection().entrySet().iterator();
                while (it.hasNext()) {
                    Map.Entry me = (Map.Entry) it.next();
                    PrimitiveElement pe =
                            gcr.getHiddenField((String) me.getValue());
                    String colname;
                    if (pk2fk == null) {
                        // default mapping
                        colname = pe.getName();
                    } else {
                        RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
                                trcm.getPrimitiveElementMapping(
                                        (String) tclassndproj.get(me.getKey()));
                        colname = (String) pk2fk.get(pem.getName());
                    }
                    table.createPrimitiveElementMapping(pe, colname);
                }
            }
        }
    }

    /**
     * Calculates the main table name of a Class. First it looks in the meta
     * information to find if it is not already defined. If it is not defined
     * then the table name is built with the class name (not fully qualified).
     */
    private String getTableName(SpeedoClass clazz) {
        return clazz.name;
    }

    private String getTableName(GenClassRef gcr) {
        String t = gcr.getGenClassId();
        int idx = t.lastIndexOf('.');
        if (idx != -1) {
            t = t.substring(idx + 1);
        }
        return t.replace(JormPathHelper.SEP, '_');
    }

    private String getColumnName(PrimitiveElement pe) {
        String c = pe.getName();
        int idx = c.lastIndexOf('.');
        if (idx != -1) {
            c = c.substring(idx +1);
        }
        return c;
    }

    private Map getPk2Fk(String fks) {
        Map res = new HashMap();
        StringTokenizer st = new StringTokenizer(fks, "=,:;/", false);
        String pk = null;
        while(st.hasMoreTokens()) {
            String tok = st.nextToken();
            tok = tok.trim();
            if (pk == null) {
                pk = tok;
            } else {
                res.put(pk, tok);
                pk = null;
            }
        }
        return res;
    }

    private Manager getManager(MetaObject mo) {
        MetaObject m = mo;
        while(m != null && !(m instanceof Manager)) {
            m = m.getParent();
        }
        return (Manager) m;
    }

    private boolean isColocatedGCTable(SpeedoField sf) {
        return sf.getExtension(VENDOR_NAME, JOIN_TABLE) == null
                &&  sf.getExtension(VENDOR_NAME, TARGET_FK) == null
                &&  sf.getExtension(VENDOR_NAME, SOURCE_FK) != null;
    }
}
