/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 0.1
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: P. Dechamboux, Y. Devaux, F. Dhumeaux.
 *
 */

package org.objectweb.speedo.metadata;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Collection;
import java.io.File;

/**
 * Describes a persistence capable class.
 * @author	Y. Devaux
 */
public class SpeedoClass extends SpeedoElement {
    /**
     * Class name.
     */
    public String name;

    /**
     * Attribute identity-type.
     */
    public byte identityType = SpeedoIdentity.CONTAINER_ID;

    /**
     * Attribute objectid-class.
     */
    public String objectidClass;

    /**
     * Attribute requires-extent.
     */
    public boolean requiresExtent = true;

    /**
     * Super class name.
     */
    public String persistenceCapableSuperClassName = "";

    /**
     * Description of persistent capable fields of this class. The HashMap key
     * is the field name.
     */
    public Map jdoField = new HashMap();

    /**
     * Descriptor of this class code.
     */
    public ClassCode classCode = new ClassCode();

    /**
     * Description of the package which contains this class.
     */
    public SpeedoPackage jdoPackage;

    /**
     * Indicates if the class is abstract.
     */
    public boolean isAbstract = false;

    /**
     * Indicates if the class implements InstanceCallbacks
     */
    public boolean isInstanceCallbacks = false;

    /**
     * Indicates if the class implements InstanceCallbacks
     */
    public boolean isSerializable = false;

    /**
     * Class signature.
     */
    public String signature;

    /**
     * Version UID for serialization.
     */
    public long VersionUID = 0;

    /**
     * Indicates whether something goes wrong during the JDO Instances generation.
     */
    public boolean failed = false;

    /**
     * Class jorm corresponding to this jdoclass
     */
    public Class jormclass;

    /**
     * Number of fields of this class (including inherited fields).
     */
    private int fieldsCount = -1;

    private int pkFieldsCount = -1;

    public String getFQName() {
        if (jdoPackage.name == null || jdoPackage.name.length() == 0) {
            return name;
        } else {
            return jdoPackage.name + "." + name;
        }
    }

    /**
     * Transforms a SpeedoClass into a String.
     * @return the Sting corresponding to the SpeedoClass.
     */
    public String toString() {
        String s = ("\n class name : " + name +
                ", identityType : " + identityType +
                ", objectidClass : " + objectidClass +
                ", requiresExtent : " + requiresExtent +
                ", persistenceCapableSuperClass : " + persistenceCapableSuperClassName);
        Iterator it = jdoField.values().iterator();
        while (it.hasNext()) {
            s = s + "\t" + it.next().toString();
        }
        return s;
    }

    /**
     * Adds a SpeedoField to the class.
     * Precondition: this field doesn't exist in the jdoFields HashMap.
     * @param field field to add.
     */
    public void add(Object field) {
        SpeedoField f = (SpeedoField) field;
        f.jdoClass = this;
        jdoField.put(f.name, field);
    }

    /**
     * Adds a SpeedoField to the class.
     * @param field field to add.
     * @param failsOnError if an error must be thrown or creates a warning.
     * @param logger logger for writting warn message if necessary.
     * @exception SpeedoException if the field was already defined into the class.
     */
    public void add(Object field, boolean failsOnError, Logger logger) throws SpeedoException {
        SpeedoField f = (SpeedoField) field;
        if (jdoField.containsKey(f.name)) {
            if (failsOnError)
                throw new SpeedoException("The field " + f.name + " of class "
                        + name + " is defined twice.");
            else
                logger.log(BasicLevel.ERROR, "The field " + f.name + " of class "
                        + name + " is defined twice.");
        } else {
            f.jdoClass = this;
            jdoField.put(f.name, field);
        }
    }

    /**
     * Computes the field numbers for the fields of this class.
     *
     * @return the number of fields of this class (including inherited fields).
     */
    public int computeFieldNumbers() {
        if (fieldsCount != -1) {
            return fieldsCount;
        }
        if (persistenceCapableSuperClassName.length() == 0) {
            fieldsCount = 0;
        } else {
            // computes the field numbers of the super class
            int p = persistenceCapableSuperClassName.lastIndexOf('.');
            String pkg;
            String cls;
            if (p == -1) {
                pkg = "";
                cls = persistenceCapableSuperClassName;
            } else {
                pkg = persistenceCapableSuperClassName.substring(0, p);
                cls = persistenceCapableSuperClassName.substring(p + 1);
            }

            SpeedoPackage spkg = (SpeedoPackage) jdoPackage.
                    jdoXMLDescriptor.jdoPackage.get(pkg);
            SpeedoClass jdoSuperClass = (SpeedoClass) spkg.jdoClass.get(cls);

            fieldsCount = jdoSuperClass.computeFieldNumbers();
        }
        Iterator i = jdoField.values().iterator();
        while (i.hasNext()) {
            SpeedoField jdoField = (SpeedoField) i.next();
            jdoField.number = fieldsCount++;
        }
        return fieldsCount;
    }

    public String getJormFileName() {
        return getFQName().replace('.', File.separatorChar) + ".pd";
    }

    public boolean generateObjectId() {
        return getPkFieldCount() > 1;
    }

    public int getPkFieldCount() {
        if (pkFieldsCount == -1) {
            pkFieldsCount = 0;
            if (identityType == SpeedoIdentity.USER_ID &&
                    (objectidClass == null || objectidClass.length() == 0)) {
                Iterator it = jdoField.values().iterator();
                while(it.hasNext() && pkFieldsCount < 2) {
                    if (((SpeedoField) it.next()).primaryKey) {
                        pkFieldsCount++;
                    }
                }
            }
        }
        return pkFieldsCount;
    }
}
