/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.runtime;

import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.perseus.persistence.api.TransactionalPersistenceManager;
import org.objectweb.perseus.persistence.lib.BasicWorkingSet;
import org.objectweb.speedo.api.ProxyManager;
import org.objectweb.speedo.api.Transaction;
import org.objectweb.speedo.api.WorkingSetLifeCycle;
import org.objectweb.speedo.api.ExceptionHelper;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.fractal.api.control.LifeCycleController;

import javax.jdo.JDOException;
import javax.jdo.PersistenceManager;
import javax.transaction.Synchronization;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class SpeedoTransaction
        extends BasicWorkingSet
        implements Transaction, LifeCycleController {

    public final static String PROXY_MANAGER_BINDING = "proxy-manager";
    public final static String TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING
            = "transactional-persistence-manager";

    private ProxyManager pm = null;
    private TransactionalPersistenceManager tpm = null;

    private boolean nontransactionalRead;
    private boolean nontransactionalWrite;
    private boolean retainValues;
    private boolean restoreValues;
    private boolean optimistic;
    private Synchronization synchronization = null;
    private Object dataAccess = null;
    private boolean managedEnv = false;


    /**
     * It activates the working set. This is used to delimit the begining of
     * the working set.
     */
    public void active() throws PersistenceException {
        status = WorkingSetLifeCycle.getNextStatus(
                status, WorkingSetLifeCycle.ACTIVE_ACTION);
    }

    public void assignDataAccess(Object dataaccess) {
        this.dataAccess = dataaccess;
    }

    public Object getDataAccess() {
        return dataAccess;
    }

    public void removeDataAccess(Object dataaccess) {
        dataAccess = null;
    }

    public boolean isManagedEnv() {
        return managedEnv;
    }

    // IMPLEMENTATION OF THE LifeCycleController INTERFACE //
    //-----------------------------------------------------//

    public String getFcState() {
        return null;
    }

    public void startFc() {
        managedEnv = pm.getPersistenceManagerFactory().getProperties()
            .getProperty(SpeedoProperties.MANAGED, "")
            .equals("true");
    }

    public void stopFc() {
    }

    // IMPLEMENTATION OF THE UserBindingController INTERFACE //
    //-------------------------------------------------------//

    public Object getFcBindings(String c) {
        if (PROXY_MANAGER_BINDING.equals(c))
            return pm;
        else if (TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING.equals(c))
            return tpm;
        else
            return super.getFcBindings(c);
    }

    public void addFcBinding(String c, Object s) {
        if (PROXY_MANAGER_BINDING.equals(c))
            pm = (ProxyManager) s;
        else if (TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING.equals(c))
            tpm = (TransactionalPersistenceManager) s;
        else
            super.addFcBinding(c, s);
    }

    public void removeFcBinding(String c, Object s) {
        if (PROXY_MANAGER_BINDING.equals(c))
            pm = null;
        else if (TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING.equals(c))
            tpm = null;
        else
            super.removeFcBinding(c, s);
    }


    // IMPLEMENTATION OF THE javax.jdo.Transactionn INTERFACE //
    //--------------------------------------------------------//

    public void begin() {
        try {
            tpm.begin(this);
        } catch (PersistenceException e) {
            throw new JDOException("", ExceptionHelper.getNested(e));
        }
    }

    public void commit() {
        logger.log(BasicLevel.DEBUG, "Commit the transaction");
        if (synchronization != null)
            synchronization.beforeCompletion();
        try {
            tpm.commit(this);
        } catch (PersistenceException e) {
            throw new JDOException("", ExceptionHelper.getNested(e));
        }
    }

    public void rollback() {
        try {
            tpm.rollback(this);
        } catch (PersistenceException e) {
            throw new JDOException("", ExceptionHelper.getNested(e));
        }
    }

    public boolean isActive() {
        return status == TransactionalPersistenceManager.CTX_ACTIVE_TRANSACTIONAL;
    }

    public void setNontransactionalRead(boolean b) {
        nontransactionalRead = b;
    }

    public boolean getNontransactionalRead() {
        return nontransactionalRead;
    }

    public void setNontransactionalWrite(boolean b) {
        nontransactionalWrite = b;
    }

    public boolean getNontransactionalWrite() {
        return nontransactionalWrite;
    }

    public void setRetainValues(boolean b) {
        retainValues = b;
    }

    public boolean getRetainValues() {
        return retainValues;
    }

    public void setRestoreValues(boolean b) {
        restoreValues = b;
    }

    public boolean getRestoreValues() {
        return restoreValues;
    }

    public void setOptimistic(boolean b) {
        optimistic = b;
    }

    public boolean getOptimistic() {
        return optimistic;
    }

    public void setSynchronization(Synchronization s) {
        synchronization = s;
    }

    public Synchronization getSynchronization() {
        return synchronization;
    }

    public PersistenceManager getPersistenceManager() {
        return pm;
    }
}
