/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.runtime.genclass;

import org.objectweb.speedo.api.SpeedoGenClassProxy;
import org.objectweb.speedo.api.LifeCycle;
import org.objectweb.speedo.api.ProxyManagerSwitch;
import org.objectweb.speedo.api.Transaction;
import org.objectweb.speedo.api.SpeedoAccessor;
import org.objectweb.speedo.api.SpeedoProxy;
import org.objectweb.speedo.api.ProxyManager;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PAccessor;
import org.objectweb.jorm.api.PIndexedElem;
import org.objectweb.jorm.api.PBindingCtrl;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.perseus.persistence.api.TransactionalPersistenceManager;
import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.perseus.cache.api.UnFixProtocolException;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.LoggerFactory;

import javax.jdo.PersistenceManagerFactory;
import javax.jdo.PersistenceManager;
import javax.jdo.JDOUnsupportedOptionException;
import javax.jdo.JDOException;
import javax.jdo.spi.PersistenceCapable;
import javax.jdo.spi.StateManager;
import java.util.Collection;
import java.util.Iterator;

/**
 * This class is a basic implementation of the SpeedoGenClassProxy interface.
 * It uses a delegate PBinding.
 *
 * @author S.Chassande-Barrioz
 */
public abstract class GenClass
        implements SpeedoGenClassProxy, Loggable, PBindingCtrl {

    /**
     * This fields indicates if the proxy is active (attached to at least a
     * Proxy Manager)
     */
    protected boolean jdoIsActive = false;

    /**
     * The status of the instance used in pessimistic mode or outside a
     * transaction. Otherwise the $classNameFields contains its status.
     */
    protected byte jdoStatus = LifeCycle.initState();

    /**
     * This is field is null when the proxy is not active
     */
    protected PersistenceManagerFactory pmf = null;

    /**
     * This is field is null when the proxy is not active
     */
    protected ProxyManagerSwitch pms = null;

    /**
     * This field is the Transactional persistence manager.
     */
    protected TransactionalPersistenceManager tpm = null;

    /**
     * It is the list of accessors (optimistic case)
     */
    protected Collection accessors = null;

    /**
     * This field is the transaction (Working set associate to the proxy used
     * in pessimistic mode or outside transaction.
     */
    protected Transaction tx = null;

    /**
     * This field counts the fix operation done over this ReplaceableCacheEntry
     */
    protected int fixCount = 0;

    /**
     * This field is a logical clock used by the ReplacementManager
     */
    protected long age = 0;

    /**
     * is the PBinding which permits to store/load the data into/from the support
     */
    protected PBindingCtrl pbinding;

    /**
     * Accessor used in pessimistic and persistent non transactional mode
     */
    protected SpeedoAccessor accessor;

    /**
     * The logger used to trace this generic class
     */
    protected Logger logger;

    /**
     * is the name of the mapper inside which the generic class is persistent
     */
    protected String mapperName;

    /**
     * is the name of the project name inside which the generic class is used.
     */
    protected String projectName;

    /**
     * is the type of the generic class
     */
    protected PType ptype;

    /**
     * TODO
     */
    protected String linkedField = null;

    /**
     * Is the PName which
     */
    protected Object pnameHints = null;

    protected Object inner = null;

    public GenClass() {
    }

    public GenClass(PType ptype,
                    String linkedField,
                    Object pnameHints,
                    String mapperName) {
        this.pnameHints = pnameHints;
        this.ptype = ptype;
        this.mapperName = mapperName;
        this.linkedField = linkedField;
    }

    public boolean elementIsReference() {
        return ptype.getNestedPType().getTypeCode() == PType.TYPECODE_REFERENCE;
    }

    public void jdoMakePersistent(ProxyManager pm, SpeedoProxy sp) {
        if (!sp.jdoIsActive()) {
            if (sp instanceof SpeedoGenClassProxy) {
                SpeedoGenClassProxy sgcp = (SpeedoGenClassProxy) sp;
                sgcp.jdoSetMapperName(jdoGetMapperName());
                if (sgcp.jdoGetPType() == null) {
                    //Assign the PType to the generic class
                    sgcp.jdoSetPType(jdoGetPType().getNestedPType());
                }
                sgcp.jdoSetLinkedField(jdoGetGenClassId());
                sgcp.jdoSetPNameHints(getPName());
            }
            pm.makePersistent(sp);
        }
    }

    // ABSTRACT METHODS //
    // -----------------//

    protected abstract SpeedoAccessor createAccessor();

    public abstract void jdoLoadFieldsInAccessor(SpeedoAccessor sa);

    public abstract void jdoLoadFieldsFromAccessor(SpeedoAccessor sa);

    public abstract Collection getGenClassElements(SpeedoAccessor sa);

    public abstract void setGCValue(Object o);

    // IMPLEMENTATION OF THE PBindingCtrl INTERFACE //
    // ---------------------------------------------//

    public void bind(PName name) throws PException {
        pbinding.bind(name);
    }

    public boolean exist(Object o) throws PException {
        return pbinding.exist(o);
    }

    public PName export(Object o) throws PException {
        return pbinding.export(o);
    }

    public PName export(Object o, Object o1) throws PException {
        return pbinding.export(o, o1);
    }

    public PClassMapping getPClassMapping() {
        return pbinding.getPClassMapping();
    }

    public PName getPName() {
        return pbinding.getPName();
    }

    public byte getStatus() {
        return pbinding.getStatus();
    }

    public void setPName(PName pName) {
        pbinding.setPName(pName);
    }

    public void setStatus(byte b) {
        pbinding.setStatus(b);
    }

    public void init(PClassMapping mapping) throws PException {
        pbinding.init(mapping);
    }

    public void read(Object o, PAccessor accessor) throws PException {
        pbinding.read(o, accessor);
    }

    public void read(Object o, PAccessor pAccessor, Object o1) throws PException {
        pbinding.read(o, pAccessor, o1);
    }

    public void unbind() throws PException {
        pbinding.unbind();
    }

    public void unexport(Object o) throws PException {
        pbinding.unexport(o);
    }

    public void write(Object o, PAccessor accessor) throws PException {
        if (logger != null)
            logger.log(BasicLevel.DEBUG, "Writing the gen class: " + linkedField);
        if (inner != null) {
            setGCValue(inner);
        }
        pbinding.write(o, accessor);
        GenClassAccessor gca = (GenClassAccessor) accessor;
        for(Iterator i = getGenClassElements(gca).iterator();i.hasNext();) {
            GenClassElement gce = (GenClassElement) i.next();
            if (gce.getElemStatus() != PIndexedElem.ELEM_DELETED) {
                gce.setStatus(PIndexedElem.ELEM_UNMODIFIED);
            } else {
                // remove the element
                i.remove();
            }
        }
        //Re initialiaze the flag after the elements have been written.
        gca.supportDelta = true;
    }


    // IMPLEMENTATION OF THE PersistentCapable INTERFACE //
    //---------------------------------------------------//
    // Speedo implements partialy  the javax.jdo.spi package. Then only the
    // minimal method are implemented. The others throw a
    // JDOUnsupportedOptionException exception.

    public PersistenceManager jdoGetPersistenceManager() {
        return (jdoIsActive ? pms.lookup(pmf) : null);
    }

    public void jdoReplaceFlags() {
        throw new JDOUnsupportedOptionException();
    }

    public PersistenceCapable jdoNewInstance(StateManager sm) {
        throw new JDOUnsupportedOptionException();
    }

    public PersistenceCapable jdoNewInstance(StateManager sm, Object oid) {
        throw new JDOUnsupportedOptionException();
    }

    public Object jdoNewObjectIdInstance() {
        throw new JDOUnsupportedOptionException();
    }

    public Object jdoGetObjectId() {
        return getPName();
    }

    public Object jdoGetTransactionalObjectId() {
        return getPName();
    }

    public void jdoReplaceField(int fieldNumber) {
        throw new JDOUnsupportedOptionException();
    }

    public void jdoReplaceFields(int[] fieldNumbers) {
        throw new JDOUnsupportedOptionException();
    }

    public void jdoProvideField(int fieldNumber) {
        throw new JDOUnsupportedOptionException();
    }

    public void jdoProvideFields(int[] fieldNumbers) {
        throw new JDOUnsupportedOptionException();
    }

    public void jdoCopyFields(Object pc, int[] fieldNumbers) {
        throw new JDOUnsupportedOptionException();
    }

    public void jdoMakeDirty(String fieldName) {
        // no fields => does nothing
    }

    public boolean jdoIsDirty() {
        SpeedoAccessor sa = getSpeedoAccessor();
        return sa != null
                && !LifeCycle.isTransient(sa.jdoGetStatus())
                && LifeCycle.isDirty(sa.jdoGetStatus());
    }

    public boolean jdoIsTransactional() {
        SpeedoAccessor sa = getSpeedoAccessor();
        return sa != null
                && !LifeCycle.isTransient(sa.jdoGetStatus())
                && LifeCycle.isTransactional(sa.jdoGetStatus());
    }

    public boolean jdoIsPersistent() {
        SpeedoAccessor sa = getSpeedoAccessor();
        return sa != null && LifeCycle.isPersistent(sa.jdoGetStatus());
    }

    public boolean jdoIsNew() {
        SpeedoAccessor sa = getSpeedoAccessor();
        return sa != null && LifeCycle.isNew(sa.jdoGetStatus());
    }

    public boolean jdoIsDeleted() {
        SpeedoAccessor sa = getSpeedoAccessor();
        return sa != null && LifeCycle.isDeleted(sa.jdoGetStatus());
    }

    public void jdoReplaceStateManager(StateManager sm) {
        // nothing to do, No State manager is used
    }

    public Object jdoNewObjectIdInstance(String s) {
        return null;
    }

    public void jdoCopyKeyFieldsToObjectId(Object o) {
        throw new JDOUnsupportedOptionException();
    }

    public void jdoCopyKeyFieldsToObjectId(PersistenceCapable.ObjectIdFieldSupplier objectIdFieldSupplier, Object o) {
        throw new JDOUnsupportedOptionException();
    }

    public void jdoCopyKeyFieldsFromObjectId(PersistenceCapable.ObjectIdFieldConsumer objectIdFieldConsumer, Object o) {
        throw new JDOUnsupportedOptionException();
    }


    // IMPLEMENTATION OF THE SpeedoGenClassProxy INTERFACE //
    //-----------------------------------------------------//

    /**
     * @param mn is the mapper name by which the instance is persistent.
     */
    public void jdoSetMapperName(String mn) {
        mapperName = mn;
    }

    /**
     * Assignes the PType of the generic class.
     * @param ptype of the generic class
     */
    public void jdoSetPType(PType ptype) {
        this.ptype = ptype;
    }

    /**
     * @return the PType of the generic class.
     */
    public PType jdoGetPType() {
        return ptype;
    }

    /**
     * assignes the PBinding to use to load/store the generic class
     * Call this method with null value means that the generic is no more
     * persistent.
     * @param pb is the PBinding.
     */
    public void jdoSetPBinding(PBinding pb) {
        pbinding = (PBindingCtrl) pb;
    }

    /**
     * @return the pbinding associated to the generic class. The value is if the
     * generic class is not persistent.
     */
    public PBinding jdoSetPBinding() {
        return pbinding;
    }

    /**
     * @return a string which represents the fully path of the persistent field
     * of the class with wich the generic class is persistent.
     * If the linked fiekf is the 'f1' provided by the class 'com.foo.Bar'
     * then the returned string will be 'com.foo.Bar/f1'
     */
    public String jdoGetGenClassId() {
        return linkedField;
    }

    /**
     * Assignes the linked field name for this generic class.
     * @param lf is tring which represents the fully path of the persistent
     * field of the class with wich the generic class is persistent.
     * If the linked fiekf is the 'f1' provided by the class 'com.foo.Bar'
     * then the expected string is 'com.foo.Bar/f1'
     */
    public void jdoSetLinkedField(String lf) {
        linkedField = lf;
    }

    /**
     * Assignes the new value of the generic class (Collection, Map or Array
     * or ...).
     */
    public void setInner(Object o) {
        inner = o;
    }
    // IMPLEMENTATION OF THE SpeedoProxy INTERFACE //
    //---------------------------------------------//

    public Object jdoGePNameHints() {
        return pnameHints;
    }

    public void jdoSetPNameHints(Object hints) {
        pnameHints = hints;
    }

    public String jdoGetMapperName() {
        return mapperName;
    }

    public String jdoGetProjectName() {
        return projectName;
    }


    public boolean jdoIsActive() {
        return jdoIsActive;
    }

    public void jdoIsActive(boolean newvalue) {
        //System.out.println("jdoIsActive(" + newValue + ") old=" + jdoIsActive + " / this=" + this);
        jdoIsActive = newvalue;
    }

    public SpeedoAccessor getSpeedoAccessor() {
        if (!jdoIsActive) {
            return this;
        }
        logger.log(BasicLevel.DEBUG, "active");
        Transaction tx = (Transaction) pms.lookup(pmf).currentTransaction();
        logger.log(BasicLevel.DEBUG, "tx = " + tx);

        if (tx.isActive() && tx.getOptimistic()) {
            logger.log(BasicLevel.DEBUG, "tx active & optim");
            SpeedoAccessor sa = (SpeedoAccessor) tx.lookup(getPName());
            if (sa == null) {
                sa = createAccessor();
                jdoLoadFieldsInAccessor(sa);
                sa.jdoSetSpeedoProxy(this);
                accessors.add(sa);
            }
            return sa;
        } else {
            logger.log(BasicLevel.DEBUG, "return pessimistic accessor");
            if (accessor == null) {
                accessor = createAccessor();
                jdoLoadFieldsInAccessor(accessor);
                accessor.jdoSetSpeedoProxy(this);
            }
            return accessor;
        }
    }

    public Collection getSpeedoAccessors() {
        return accessors;
    }

    public void removeSpeedoAccessor(SpeedoAccessor sa) {
        accessors.remove(sa);
    }

    /**
     * The identifier of a generic class is based on the identifier of
     * the class which references this generic class.
     * @return false;
     */
    public boolean jdoIsContainerId() {
        return pnameHints == null;
    }

    public ProxyManagerSwitch jdoGetPersistenceManagerSwitch() {
        return pms;
    }

    public void jdoSetPersistenceManagerSwitch(ProxyManagerSwitch pms) {
        this.pms = pms;
    }

    public PersistenceManagerFactory jdoGetPersistenceManagerFactory() {
        return pmf;
    }

    public void jdoSetPersistenceManagerFactory(PersistenceManagerFactory pmf) {
        this.pmf = pmf;
    }

    public TransactionalPersistenceManager jdoGetTransactionalPersistenceManager() {
        return tpm;
    }

    public void jdoSetTransactionalPersistenceManager(TransactionalPersistenceManager tpm) {
        this.tpm = tpm;
    }

    public void jdoReadIntention(SpeedoAccessor sa) {
        try {
            tpm.readIntention(sa.jdoGetTransaction(), sa.getCeIdentifier());
        } catch (PersistenceException e) {
            throw new JDOException("Impossible to notify a read intention", e);
        }
    }

    public void jdoWriteIntention(SpeedoAccessor sa) {
        try {
            Transaction tx = sa.jdoGetTransaction();
            tpm.readIntention(tx, getPName());
            tpm.writeIntention(tx, sa);
        } catch (PersistenceException e) {
            throw new JDOException("Impossible to notify a write intention", e);
        }
    }

    public Object getJormConfig() {
//TODO
        return null;
    }

    public String jdoGetJdoFileName() {
        return "${jdoXmlDescFN}"; // TODO
    }

    public void jdoDeletePersistent(ProxyManager pm) {
        //Nothing to do
    }

    // IMPLEMENTATION OF THE PAccessor INTERFACE //
    //-------------------------------------------//

    public Object getMemoryInstance() {
        return this;
    }

    // IMPLEMENTATION OF THE SpeedoAccessor INTERFACE //
    //------------------------------------------------//

    public void jdoChangeStatus(byte action) {
        if (!jdoIsActive) {
            LifeCycle.nextStatePersistenceCapable(jdoStatus, action);
        } else {
            getSpeedoAccessor().jdoChangeStatus(action);
        }
    }

    public byte jdoGetStatus() {
        if (!jdoIsActive) {
            return jdoStatus;
        } else {
            return getSpeedoAccessor().jdoGetStatus();
        }
    }

    public void jdoSetStatus(byte newValue) {
        if (!jdoIsActive) {
            jdoStatus = newValue;
        } else {
            getSpeedoAccessor().jdoSetStatus(newValue);
        }
    }

    public SpeedoProxy jdoGetSpeedoProxy() {
        return this;
    }

    public void jdoSetSpeedoProxy(SpeedoProxy sp) {
    }

    public Transaction jdoGetTransaction() {
        return tx;
    }

    public void jdoSetTransaction(Transaction tx) {
        this.tx = tx;
    }

    // IMPLEMENTATION OF THE CacheEntry INTERFACE //
    //-------------------------------------------//

    public Object getCeObject() {
        return this;
    }

    public Object getCeIdentifier() {
        return getPName();
    }


    // IMPLEMENTATION OF THE FixableCacheEntry INTERFACE //
    //---------------------------------------------------//

    public void fixCe() {
        fixCount++;
    }

    public void unfixCe() throws UnFixProtocolException {
        fixCount++;
    }

    public int getCeFixCount() {
        return fixCount;
    }

    // IMPLEMENTATION OF THE ReplaceableCacheEntry INTERFACE //
    //-------------------------------------------------------//

    public long getCeAge() {
        return age;
    }

    public void setCeAge(long l) {
        age = l;
    }

    // ------------------------------------------------------------------------
    // IMPLEMENTATION OF THE Loggable INTERFACE
    // ------------------------------------------------------------------------

    public void setLogger(Logger _logger) {
        logger = _logger;
    }

    public void setLoggerFactory(LoggerFactory loggerfactory) {
        logger = loggerfactory.getLogger(this.getClass().getName());
    }

    public Logger getLogger() {
        return logger;
    }

    public LoggerFactory getLoggerFactory() {
        return null;
    }
}
