/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.runtime.mim;

import org.objectweb.perseus.persistence.api.MemoryInstanceManager;
import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.api.PException;

import javax.jdo.spi.JDOImplHelper;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Collections;
import java.util.StringTokenizer;

/**
 * This class is an implementation of the MemoryInstanceManager provided by the
 * persistence framwork availlable in perseus. This implementation depends on
 * the identifier used, and the way to instanciate user objects. This
 * implementation supports only the jorm object identifier: PName.
 * This implementation supports also the instanciation of the Jorm generic
 * class. To do this it is needed to specify for each jorm generic class used
 * the name of the class which must be instanciated. This configuration is done
 * via the GenClassNames attribute.
 * This implementation does not used a pool, then the instance are created at
 * each newInstance call.
 *
 * @author S.Chassande-Barrioz
 */
public class SpeedoMemoryInstanceManager
        implements MemoryInstanceManager, SpeedoMemoryInstanceManagerAttribute {

    /**
     * This fields maintains the association between the name of the Jorm
     * generic class and the class name which must be implemented.
     */
    private Map name2gcimpl = null;


    public SpeedoMemoryInstanceManager() {
        name2gcimpl = Collections.EMPTY_MAP;
    }

    // IMPLEMENTATION OF THE SpeedoMemoryInstanceManagerAttribute INTERFACE //
    //----------------------------------------------------------------------//

    /**
     * @return a String describing the genclass names with the following format:
     * "(jorm_name,java_name),(jorm_name,java_name),(jorm_name,java_name)}"
     */
    public String getGenClassNames() {
        if (name2gcimpl == null || name2gcimpl.size() == 0)
            return "{}";
        StringBuffer res = new StringBuffer("");
        boolean first = true;
        for (Iterator it = name2gcimpl.entrySet().iterator(); it.hasNext();) {
            if (first) {
                res.append("{");
                first = false;
            } else
                res.append(",");
            res.append("(");
            Map.Entry me = (Map.Entry) it.next();
            res.append(me.getKey());
            res.append(",");
            res.append(me.getValue());
            res.append(")");
        }
        res.append("}");
        return res.toString();
    }

    /**
     * It assignes a description of the gen class names.
     * @param gcname is the desciption which must follow this format:
     * "(jorm_name,java_name),(jorm_name,java_name),(jorm_name,java_name)}"
     */
    public void setGenClassNames(String gcname) {
        if (gcname == null || gcname.length() == 0)
            name2gcimpl = Collections.EMPTY_MAP;
        name2gcimpl = new HashMap(5);
        StringTokenizer st = new StringTokenizer(gcname, "{,()}:;[]", false);
        String key = null;
        while (st.hasMoreTokens()) {
            if (key == null) {
                key = st.nextToken().trim();
            } else {
                name2gcimpl.put(key, st.nextToken().trim());
            }
        }
    }


    // IMPLEMENTATION OF THE MemoryInstanceManager INTERFACE //
    //-------------------------------------------------------//

    /**
     * It creates an instance since an identifier
     * @param oid is the identifier of the futur object
     * @return a memory instance
     */
    public Object newInstance(Object oid, Object context) throws PersistenceException {
        if (!(oid instanceof PName))
            throw new PersistenceException("Unmanaged object identifier: " + oid);
        PName pn = (PName) oid;
        String className = null;
        try {
            pn = pn.resolve(context);
            String jormName = pn.getPType().getJormName();
            className = (String) name2gcimpl.get(className);
            if (className == null) {
                // The PName represents a user object
                className = jormName;
                return Class.forName(className).newInstance();
            } else {
                // The PName represents a generic jorm class
                return Class.forName(className).newInstance();
            }
        } catch (PException e) {
            throw new PersistenceException(
                    "Impossible to analyze the object identifier" + oid, e);
        } catch (Exception e) {
            throw new PersistenceException(
                    "Impossible to instanciate the class " + className
                    + "Be careful to the generic class (Collection, ...)", e);
        }

    }

    /**
     * It notify that an instance is no more used.
     * @param obj
     */
    public void releaseInstance(Object obj) {
        //Nothing to do
    }
}
