/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 0.1
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: X. Spengler
 *
 */

package org.objectweb.speedo.runtime.query;

import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.api.PMappingStructuresManager;
import org.objectweb.jorm.api.PMapperListener;
import org.objectweb.jorm.api.PMapCluster;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.runtime.JormPathHelper;
import org.objectweb.jorm.xml2mi.api.Parser;
import org.objectweb.medor.api.EvaluationException;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.eval.api.ConnectionResources;
import org.objectweb.medor.eval.api.QueryEvaluator;
import org.objectweb.medor.eval.lib.MedorEvaluator;
import org.objectweb.medor.eval.prefetch.api.PrefetchCache;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.filter.lib.BasicParameterOperand;
import org.objectweb.medor.optim.api.QueryTransformer;
import org.objectweb.medor.optim.jorm.JormFlatten2Rdb;
import org.objectweb.medor.optim.lib.IndexesGenerator;
import org.objectweb.medor.optim.lib.FlattenQueryTreeRule;
import org.objectweb.medor.optim.lib.BasicQueryRewriter;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.OrderField;
import org.objectweb.medor.query.jorm.api.JormExtent;
import org.objectweb.medor.query.jorm.lib.JormQueryTreeHelper;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.medor.query.lib.SelectProject;
import org.objectweb.medor.query.lib.Project;
import org.objectweb.medor.query.lib.BasicOrderField;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.type.lib.PTypeSpaceMedor;
import org.objectweb.perseus.pool.api.PoolResource;
import org.objectweb.perseus.pool.api.Pool;
import org.objectweb.speedo.api.JormFactory;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.api.ProxyManager;
import org.objectweb.speedo.runtime.query.api.QueryCompiler;
import org.objectweb.speedo.runtime.query.api.QueryDefinition;
import org.objectweb.speedo.runtime.query.parser.ASTSpeedoQL;
import org.objectweb.speedo.runtime.query.parser.ParseException;
import org.objectweb.speedo.runtime.query.parser.SimpleNode;
import org.objectweb.speedo.runtime.query.parser.SpeedoQL;
import org.objectweb.speedo.runtime.query.parser.SpeedoQLQueryFilterVisitor;
import org.objectweb.speedo.runtime.query.parser.SpeedoQLVariableVisitor;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import javax.jdo.Extent;
import javax.jdo.PersistenceManager;
import java.io.CharArrayReader;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.ArrayList;
import java.util.List;
import java.util.HashMap;

/**
 * SpeedoQueryCompiler object represents a query. This object is created
 * when a new query is created, and can be used several times. A list of
 * SpeedoQueryCompiler is managed with a SpeedoQueryManager object.
 * When a user creates a new JDO Query object (SpeedoQuery), a
 * SpeedoQueryCompiler object is associated to the SpeedoQuery object which is
 * used to delegate some methods.
 */
public class SpeedoQueryCompiler
        implements QueryCompiler, PoolResource {

    public final static String LOGGER_NAME
            = "org.objectweb.speedo.rt.query.SpeedoQueryCompiler";
    public final static String PARSER_LOGGER_NAME
            = "org.objectweb.speedo.rt.query.parser";
    public final static String PARSER_VAR_LOGGER_NAME
            = PARSER_LOGGER_NAME + ".variable";
    public final static String PARSER_FILTER_LOGGER_NAME
            = PARSER_LOGGER_NAME + ".filter";

    private short status;

    /**
     * Fields from javax.jdo.Query
     */
    private java.lang.Class c = null;
    private String parameters = null;
    private String variables = null;
    private String filter = null;
    private List orders = null;
    private Collection candidateCollection = null;
    private Extent candidateExtent = null;

    /**
     * vparams, hparams and hvars are internal objects to manipulate the list
     * of parameters, and the list of variables.
     */
    private List vparams = null;
    private Map hparams = null;
    private Map hvars = null;
    private QueryEvaluator evaluator = null;

    /**
     * The query tree which is the result from the parsing of the given filter
     */
    private QueryTree qt = null;

    private Collection pncParams = null;
    /**
     * Logger for monolog
     */
    private Logger logger = null;
    private LoggerFactory loggerFactory = null;

    private DelegatePMapper mapper = null;
    private JormFactory jf;

    /**
     * IgnoreCache option.
     * The ignoreCache option setting specifies whether the query should execute
     * entirely in the back end, instead of in the cache.  If this flag is set
     * to true, an implementation might be able to optimize the query
     * execution by ignoring changed values in the cache.  For optimistic
     * transactions, this can dramatically improve query response times.
     *
     * IMPORTANT: for this moment, this option is not used, and the cache is not
     * used (the cache is the Object result which contains all results from the
     * previous query).
     */
    private boolean ignoreCache = false;

    private boolean prefetchResult = true;
    private PClassMapping pcm;
    private int identifierIndex = -1;

    public LoggerFactory getLoggerFactory() {
        return loggerFactory;
    }

    public void setLoggerFactory(LoggerFactory lf) {
        loggerFactory = lf;
        if (loggerFactory != null)
            logger = loggerFactory.getLogger(LOGGER_NAME);
    }

    public PMapper getMapper() {
        return mapper;
    }

    public void setMapper(PMapper m) {
        mapper = new DelegatePMapper(m);
        if (jf != null) {
            mapper.jf = jf;
        }
    }

    public void setJormFactory(JormFactory jf) {
        this.jf = jf;
        if (mapper != null) {
            mapper.jf = jf;
        }
    }

    public void defineQuery(QueryDefinition qd) throws SpeedoException {
        c = qd.getCandidateClass();
        parameters = qd.getParameters();
        variables = qd.getVariables();
        filter = qd.getFilter();
        if (qd.isCollection())
            candidateCollection = qd.getCollection();
        else
            candidateExtent = qd.getExtent();
        ignoreCache = qd.isIgnoreCache();
        orders = qd.getOrder();
        status = DEFINED;
    }

    // IMPLEMENTATION OF PoolResource INTERFACE //
    //------------------------------------------//

    public void destroy() throws Exception {
    }

    // IMPLEMENTATION OF QueryCompiler INTERFACE //
    //-------------------------------------------//

    /**
     * compile the current SpeedoQueryCompiler.
     * The query is prepared to be executed.
     * The PersistenceManager is set (even if there was a previous definition
     * of a PersistenceManager.
     */
    public void compile() throws SpeedoException, MedorException {
        if (status == UNDEFINED)
            throw new SpeedoException("Impossible to compile an undefined query");
        if (status == COMPILED)
            return;
        boolean debug = logger.isLoggable(BasicLevel.DEBUG);
        // create a speedoQL object with a filter string
        SpeedoQL sql = new SpeedoQL(new CharArrayReader(filter.toCharArray()));

        // a node is created (parsing result)
        SimpleNode node = null;
        try {
            node = sql.SpeedoQL();
        } catch (ParseException e) {
            throw new SpeedoException(
                    "Impossible to parse the filter and to create AST", e);
        }

        Manager miManager = mapper.getMetaInfoManager();

        if (miManager == null)
            throw new SpeedoException(
                    "A non null Meta information manager is needed");

        // create a new SpeedoQLVariableVisitor object with the previous
        // node and a Schema (jorm metainformation) array
        SpeedoQLVariableVisitor sqvv = new SpeedoQLVariableVisitor(
                node, miManager);
        sqvv.setLogger(loggerFactory.getLogger(PARSER_VAR_LOGGER_NAME));

        // create representations of the parameters list and the variable
        // list
        toHashtableParams(parameters, ",");
        toHashtableVars(variables, ";");

        // set these representations to the SpeedoQLVariableVisitor object
        sqvv.setParams(hparams);
        sqvv.setVars(hvars);

        // set up the nameof the current class
        // could it be null ??? to verify it ... !!!
        sqvv.setCurrentClass(c.getName());

        // start the variable visitor to catch all variables an build a
        // first tree of them
        try {
            sqvv.startVisiting();
        } catch (Exception e) {
            throw new SpeedoException(
                    "Impossible to visit the variable of the filter", e);
        }

        // get the query tree to set up the query filter visitor
        qt = sqvv.getQueryTree();

        if (debug) {
            Field[] fields = qt.getTupleStructure().getFields();
            for (int i = 0; i < fields.length; i++) {
                logger.log(BasicLevel.DEBUG,
                        "i=" + i + ", fieldname=<" + fields[i].getName() + ">");
            }
        }

        //Assign mappers and project and project name on the JormExtent nodes
        Collection extents = JormQueryTreeHelper.getJormExtents(qt);
        if (logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "Extent nodes: " + extents.size());
        try {
            for (Iterator it = extents.iterator(); it.hasNext();) {
                JormExtent je = (JormExtent) it.next();
                if (debug)
                    logger.log(BasicLevel.DEBUG, "Looking the extent of the class " + je.getJormName());
                PClassMapping pcm = jf.getPClassMapping(je.getJormName());
                if (debug)
                    logger.log(BasicLevel.DEBUG,
                            "JormExtent: " + je.getJormName() + " / pcm=" + pcm);
                je.setPMapper(pcm.getPMapper(), pcm.getProjectName());
            }
        } catch (PException e) {
            throw new SpeedoException("Error while fetching PClassPMapping", e);
        }

        // Build the root QueryNode which defines the filter and the projected
        // fields
        QueryNode root;
        if (!filter.equals("(true)") && !filter.equals("true")) {
            root = new SelectProject("", null);
            if (debug) {
                logger.log(BasicLevel.DEBUG, "filter = " + filter);
            }
            SpeedoQLQueryFilterVisitor qfilter =
                    new SpeedoQLQueryFilterVisitor(qt, (ASTSpeedoQL) node);

            qfilter.setLogger(loggerFactory.getLogger(PARSER_FILTER_LOGGER_NAME));

            // set the representations of the variables and parameters to the
            // SpeedoQLQueryFilterVisitor object
            qfilter.setParams(hparams);
            qfilter.setVars(hvars);

            // set up the nameof the current class
            // could it be null ??? to verify it ... !!!
            qfilter.setCurrentClass(c.getName());

            // start the query filter visitor, to build and expression tree of
            // the filter expression
            try {
                qfilter.startVisiting();
            } catch (Exception e) {
                throw new SpeedoException(
                        "Impossible to built the queryTree during the filter visit", e);
            }

            // this expression is the filter of the SelectProject
            Expression exprFilter = qfilter.getQueryFilter();
            if (debug) {
                logger.log(BasicLevel.DEBUG, "filter = " + exprFilter.getClass().getName());
            }
            root.setQueryFilter(exprFilter);
        } else {
            root = new Project("", null);
        }

        if (prefetchResult) {
            JormQueryTreeHelper.addPrefetchFields(sqvv.classExtent, qt, root);
            pcm = mapper.lookup(c.getName());
            identifierIndex = root.getTupleStructure().getSize() + 1;
        }

        // Project the pname
        QueryTreeField qtf = (QueryTreeField)
                qt.getTupleStructure().getField(
                        c.getName() + "." + Field.PNAMENAME);
        root.addPropagatedField(qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});

        if (orders != null && orders.size() > 0) {
            OrderField[] ofs = new OrderField[orders.size()];
            for(int i=0; i<ofs.length; i++) {
                String o = (String) orders.get(i);
                int idx = o.indexOf(' ');
                boolean desc = false;
                if (idx != -1) {
                    desc = o.substring(idx + 1).trim().equals("descending");
                    o = o.substring(0, idx);
                }
                o = c.getName() + '.' + o;
                ofs[i] = new BasicOrderField((QueryTreeField)
                    qt.getTupleStructure().getField(o), desc);
            }
            root.setOrderBy(ofs);
        }
        qt = root;

        logger.log(BasicLevel.INFO, "QueryTree built");
        if (debug) {
            QueryTreePrinter.printQueryTree(qt, logger);
        }


        // Optimize the medor QueryTree
        //-----------------------------
//TODO: remove the dependency to JormFlatten2Rdb
        ArrayList rules = new ArrayList(2);
        rules.add(new FlattenQueryTreeRule());
        rules.add(new JormFlatten2Rdb());
        QueryTransformer queryTransformer = new BasicQueryRewriter(rules);

        IndexesGenerator indexesGenerator = new IndexesGenerator();
        try {
            qt = queryTransformer.transform(qt);
            if (debug) {
                logger.log(BasicLevel.DEBUG, "Query optimized");
                QueryTreePrinter.printQueryTree(qt, logger);
            }
            qt = indexesGenerator.transform(qt);
        } catch (Exception e) {
            throw new MedorException("Impossible to optimize the query", e);
        }

        pncParams = JormQueryTreeHelper.getRequiredPNameManagers(qt);
        for (Iterator it = pncParams.iterator(); it.hasNext();) {
            ParameterOperand po = (ParameterOperand) it.next();
            String name = po.getName();
            if (debug) {
                logger.log(BasicLevel.DEBUG, "ParameterOperand " + name + " is expected.");
            }
            po.setValue(JormPathHelper.getPNameCoder(name, mapper));
        }

        // Creates an evaluator associated to the QueryTree
        evaluator = new MedorEvaluator(qt, 0);
        status = COMPILED;
        logger.log(BasicLevel.INFO, "Query compiled");
    }

    /**
     * evaluate the query with a single parameter which is a array of object
     * parameters.
     * @param a    the array parameter of the query
     * @param pm    the persistence manager object
     * @return  a Collection of result objects
     * @throws EvaluationException
     * @throws MedorException
     */
    public Collection execute(Object[] a, ProxyManager  pm)
            throws SpeedoException, MedorException {
        if (status != COMPILED)
            throw new EvaluationException(
                    "Impossible to execute a query if it has not been defined and compiled before");
        ParameterOperand[] pos = null;
        int i = 0;
        if (a != null) {
            pos = new ParameterOperand[a.length + pncParams.size()];
            for (;i < a.length; i++) {
                pos[i] = (BasicParameterOperand) hparams.get(vparams.get(i));
                pos[i].setValue(a[i]);
            }
        } else {
            pos = new ParameterOperand[pncParams.size()];
        }
        for (Iterator it = pncParams.iterator(); it.hasNext();) {
            pos[i++] = (ParameterOperand) it.next();
        }
        return executeQT(pm, pos);
    }


    /**
     * evaluate the query with a single parameter which is a Map of object parameters.
     * @param m    the map parameter of the query
     * @param pm    the persistence manager object
     * @return  a Collection of result objects
     * @throws EvaluationException
     * @throws MedorException
     */
    public Collection execute(Map m, ProxyManager  pm)
            throws SpeedoException, MedorException {
        if (status != COMPILED)
            throw new EvaluationException(
                    "Impossible to execute a query if it has not been defined and compiled before");
        ParameterOperand[] pos =
                new BasicParameterOperand[m.size() + +pncParams.size()];
        int i = 0;
        for (; i < m.size(); i++) {
            pos[i] = (BasicParameterOperand) hparams.get(vparams.get(i));
            pos[i].setValue(m.get(pos[i].getName()));
        }
        for (Iterator it = pncParams.iterator(); it.hasNext();) {
            pos[i++] = (ParameterOperand) it.next();
        }
        return executeQT(pm, pos);
    }

    public short getStatus() {
        return status;
    }

    // Private Methods //
    //-----------------//

    /**
     * Hash a String, and compute a Hashtable
     * example:
     *   ("String name, Float salary, Employee boss", ",")
     *   keys     | values
     *   ---------------------
     *   "name"   | "String"
     *   "salary" | "Float"
     *   "boss"   | "Employee"
     *
     * @param stringToHash	the String to hash
     * @param separator		the separator to tokenize
     */
    private void toHashtableParams(String stringToHash, String separator) {
        hparams = new HashMap();
        if (stringToHash != null) {
            StringTokenizer tok = new StringTokenizer(stringToHash, separator);
            vparams = new ArrayList(tok.countTokens());
            while (tok.hasMoreTokens()) {
                String tuple = tok.nextToken().trim();
                int i = tuple.indexOf(' ');
                String name = tuple.substring(i + 1, tuple.length());
                hparams.put(name, new BasicParameterOperand(getPType(tuple.substring(0, i)), name));
                vparams.add(name);
            }
        }
    }

    private void toHashtableVars(String stringToHash, String separator) {
        hvars = new HashMap();
        if (stringToHash != null) {
            StringTokenizer tok =
                    new StringTokenizer(stringToHash, separator);
            while (tok.hasMoreTokens()) {
                String tuple = tok.nextToken().trim();
                int i = tuple.indexOf(' ');
                String type = tuple.substring(0, i);
                hvars.put(tuple.substring(i + 1, tuple.length()), type);
            }
        }
    }

    /**
     * returns the PType of a given name. If the type does not exist,
     * null is returned.
     * @param name the string representation of the name.
     * @return the corresponding PType of its name.
     */
    private PType getPType(String name) {
        if ("String".equals(name)) {
            return PTypeSpace.STRING;
        }
        for(int i=0; i<PTypeSpace.PREDEFINEDPTYPES.length; i++) {
            PType type = PTypeSpace.PREDEFINEDPTYPES[i];
            if (type.getJavaName().equals(name)
                || type.getJormName().equals(name)) {
                return type;
            }
        }
        throw new RuntimeException("Not yet implemented");
        //return PTypeSpace.REFTOP;
    }

    /**
     * builds a Collection with a Medor TupleCollection.
     * @param tupleCollection   the tuple collection to read and transform into
     *                          a Collection object
     * @return  a Collection object
     */
    private Collection tupleCollection2Collection(
            TupleCollection tupleCollection,
            PersistenceManager pm,
            Object[] da,
            PMapper mapper) throws MedorException {
        if (tupleCollection == null) {
            logger.log(BasicLevel.INFO,
                    "Be careful, the result collection is empty");
        }
        QueryResultCollection qrc = new QueryResultCollection();
        qrc.init(tupleCollection, pm, da, mapper, logger);
        return qrc;
    }

    /**
     * executes a the current query, and returns a Collection of object.
     * @return  a new Collection of objects.
     * @throws EvaluationException
     * @throws MedorException
     */
    private Collection executeQT(ProxyManager pm, ParameterOperand[] pos)
            throws EvaluationException, MedorException {
        // Calculates and gets the required connection ressources for this query
        ConnectionResources connRess =
                evaluator.getRequiredConnectionResources();

        // Gets the QueryLeafs that require connections
        QueryLeaf[] leafs = connRess.getRequiredQueryLeafConnection();
        Object connection = null;
        try {
            Object cs = pm.getConnectionSpec();
            if (cs == null) {
                connection = mapper.getConnection();
            } else {
                connection = mapper.getConnection(cs);
            }
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "Problem to obtain a Connection ", e);
            throw new EvaluationException("Problem to obtain a Connection ", e);
        }

        // Setting QueryLeaf's appropriated connection Object
        for (int cpt = 0; (cpt < leafs.length); cpt++) {
            connRess.setConnection(leafs[cpt], connection);
        }
        if (prefetchResult) {
            evaluator.setPrefetchBuffer(mapper.getPrefetchCache()
                .createPrefetchBuffer(pcm, pm.currentTransaction(), identifierIndex));
        }
        // Launching Medor Evaluator...
        TupleCollection queryResult = evaluator.evaluate(pos, connRess);
        return tupleCollection2Collection(queryResult, pm, new Object[]{connection}, mapper);
    }
}
/**
 * This class is an implementation of the PMapper interface delegating all
 * method to another PMapper, except for the lookup method. In this case the
 * lookup method is done by the use of the JormFactory. The aim is to always
 * have the PClassMapping instance. But only the JormFactory is able to build
 * the PClassMapping instance of a Persistent class.
 * T
 */
class DelegatePMapper implements PMapper {

    public JormFactory jf;
    public PMapper mapper;

    public DelegatePMapper(PMapper mapper) {
        this.mapper = mapper;
    }

    // IMPLEMENTATION OF THE PMapper INTERFACE //

    public PClassMapping lookup(String s) {
        try {
            return jf.getPClassMapping(s);
        } catch (PException e) {
            return null;
        }
    }

    public void setPrefetchCache(PrefetchCache prefetchCache) throws PException {
        mapper.setPrefetchCache(prefetchCache);
    }
    public PrefetchCache getPrefetchCache() {
        return mapper.getPrefetchCache();
    }
    public void closeConnection(Object o) throws PException {
        mapper.closeConnection(o);
    }
    public String cn2mn(String s) {
        return mapper.cn2mn(s);
    }
    public Object getConnection() throws PException {
        return mapper.getConnection();
    }
    public Object getConnection(Object o) throws PException {
        return mapper.getConnection(o);
    }
    public String getMapperName() {
        return mapper.getMapperName();
    }
    public PMappingStructuresManager getPMappingStructuresManager() {
        return mapper.getPMappingStructuresManager();
    }
    public void map(PClassMapping pClassMapping) throws PException {
        mapper.map(pClassMapping);
    }
    public void map(Object o, PClassMapping pClassMapping) throws PException {
        mapper.map(o, pClassMapping);
    }
    public void map(Object o, PClassMapping pClassMapping, boolean b) throws PException {
        mapper.map(o, pClassMapping, b);
    }
    public void setConnectionFactory(Object o) throws PException {
        mapper.setConnectionFactory(o);
    }
    public Object getConnectionFactory() {
        return mapper.getConnectionFactory();
    }
    public void setMapperName(String s) {
        mapper.setMapperName(s);
    }
    public Manager getMetaInfoManager() {
        return mapper.getMetaInfoManager();
    }
    public void setMetaInfoManager(Manager manager) {
        mapper.setMetaInfoManager(manager);
    }
    public void setParser(Parser parser) {
        mapper.setParser(parser);
    }
    public void setPTypeSpace(PTypeSpace pTypeSpace) {
        mapper.setPTypeSpace(pTypeSpace);
    }
    public PTypeSpace getPTypeSpace() {
        return mapper.getPTypeSpace();
    }
    public void start() throws PException {
        mapper.start();
    }
    public void stop() throws PException {
        mapper.stop();
    }
    public void unmap(String s) throws PException {
        mapper.unmap(s);
    }
    public void addMapperEventListener(PMapperListener pMapperListener) {
        mapper.addMapperEventListener(pMapperListener);
    }
    public void removeMapperEventListener(PMapperListener pMapperListener) {
        mapper.removeMapperEventListener(pMapperListener);
    }
    public PClassMapping createGenClassMapping() {
        return mapper.createGenClassMapping();
    }
    public PMapCluster getPMapCluster(String s) throws PException {
        return mapper.getPMapCluster(s);
    }
    public Collection getPMapClusters() {
        return mapper.getPMapClusters();
    }
    public void addDependency(String s, String s1) throws PException {
        mapper.addDependency(s, s1);
    }
    public void classDefined(String s) throws PException {
        mapper.classDefined(s);
    }
}

