package org.objectweb.speedo.runtime.query.parser;

import junit.framework.TestCase;
import org.apache.tools.ant.types.DTDLocation;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.comp.compiler.lib.JormCompiler;
import org.objectweb.jorm.comp.compiler.lib.JormCompilerParameter;
import org.objectweb.jorm.comp.compiler.lib.JormCompilerConfigurator;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.util.io.lib.DirJavaExplorer;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.lib.BasicParameterOperand;
import org.objectweb.medor.filter.lib.ExpressionPrinter;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.medor.query.lib.SelectProject;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.speedo.api.ExceptionHelper;

import java.io.CharArrayReader;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Map;
import java.util.Collection;

/**
 * SpeedoQL MEDOR Test
 * Test a list of valid filters which must be parsed by MEDOR.
 */
public class SpeedoQLMedorTest extends TestCase {

    private final static String LOGGER_NAME
            = "org.objectweb.speedo.rt.query.parser.medor";

    /**
     * filterFieldParameter tests the access to simple fiels from a single
     * class, with 2 parameters (mysalary and myname)
     */
    private final static String filterThis = "(this.salary == mysalary)";

    /**
     * filterFieldParameter tests the access to simple fiels from a single
     * class, with 2 parameters (mysalary and myname)
     */
    private final static String filterFieldParameter
            = "((salary == mysalary) && (name == myname))";

    /**
     * filterFieldRef tests the access to a field from a reference. dept field
     * is a field from Employee (Department object reference, which have a name
     * field.
     */
    private final static String filterFieldRef = "(dept.name == dep)";

    /**
     * filterIn filter tests the contains method with a parameter (anEmployee)
     * with a variable: anEmployee, and a parameter: istName
     * Medor: getIn (from NavigatorFactory)
     */
    private final static String filterIn
            = "((anEmployee.name == itsName) && emps.contains(anEmployee))";

    /**
     * filterMemberOf filter tests the basic contains method with a single
     * parameter: employee (which is a parameter)
     * Not yet implemented in Medor (getMemberOf)
     */
    private final static String filterMemberOf = "(emps.contains(employee))";

    private static String projectName;
    private static String pdPath;
    private static String dtdPath;
    private static LoggerFactory loggerFactory = null;

    private static Logger logger = null;

    static {
        boolean error = false;
        projectName = System.getProperty("projectName");
        if (projectName == null) {
            System.err.println("Property projectName not set");
            error = true;
        }

        pdPath = System.getProperty("pdPath");
        if (pdPath == null) {
            System.err.println("Property pdPath not set");
            error = true;
        }

        dtdPath = System.getProperty("dtd");
        if (dtdPath == null) {
            System.err.println("Property dtdPath not set");
            error = true;
        }
        if (error)
            throw new RuntimeException("All property has not been set.");
        logger = Monolog.init().getLogger(LOGGER_NAME);
    }

    /**
     * hparams is an hashtable to store the parameters.
     * key: name of the parameter
     * value: a BasicFieldOperand
     */
    Map hparams = null;

    /**
     * hvars is an hashtable to store the variables.
     * key: name of the variable
     * value: the PType of the variable (from the declaration)
     */
    Map hvars = null;

    /**
     * builds a new SpeedoQLMedorTest
     */
    public SpeedoQLMedorTest(String name) {
        super(name);
    }

    /**
     *
     */
    public void testThis() {
        logger.log(BasicLevel.DEBUG, "SpeedoQL Test - javacc grammar");

        SpeedoQL sql = null;
        logger.log(BasicLevel.DEBUG, "Testing:" + filterThis);
        sql = new SpeedoQL(new CharArrayReader(filterThis.toCharArray()));

        try {
            ArrayList all = new ArrayList();
            all.add("org/objectweb/speedo/pobjects/ref/Employee.pd");
            all.add("org/objectweb/speedo/pobjects/ref/Department.pd");
            Manager manager = setupJorm(all);

            SimpleNode node = sql.SpeedoQL();
            SpeedoQLVariableVisitor sqvv
                    = new SpeedoQLVariableVisitor(node, manager);
            sqvv.setLogger(logger);

            toHashtableParams("Float mysalary", ",");

            sqvv.setParams(hparams);
            sqvv.setVars(null);
            sqvv.setCurrentClass("org.objectweb.speedo.pobjects.ref.Employee");

            sqvv.startVisiting();

            QueryTree queryTree = sqvv.getQueryTree();

            if (logger.isLoggable(BasicLevel.DEBUG)) {
                Field[] fields = queryTree.getTupleStructure().getFields();
                for (int i = 0; i < fields.length; i++) {
                    logger.log(BasicLevel.DEBUG, "i=" + i
                            + ", fieldname=<" + fields[i].getName() + ">");
                }
            }

            SpeedoQLQueryFilterVisitor qfilter =
                    new SpeedoQLQueryFilterVisitor(queryTree, (ASTSpeedoQL) node);
            qfilter.setLogger(logger);

            qfilter.setParams(hparams);
            qfilter.setVars(null);
            qfilter.setCurrentClass("org.objectweb.speedo.pobjects.ref.Employee");
            qfilter.startVisiting();

            // verify the expression is well-formed
            Expression expr = qfilter.getQueryFilter();
            if (expr == null)
                logger.log(BasicLevel.INFO, "ATTENTION: expr null in test");
            else
                logger.log(BasicLevel.INFO, "Expression toString... :" + ExpressionPrinter.e2str(expr));

            SelectProject sp = new SelectProject("", null);
            sp.setQueryFilter(expr);

            queryTree = sp;

            if (logger.isLoggable(BasicLevel.DEBUG)) {
                QueryTreePrinter.printQueryTree(queryTree, logger);
            }
        } catch (ParseException e) {
            throw new MyException(
                    "SpeedoQLMedorTest: Encountered errors during parse.", e);
        } catch (Exception e) {
            throw new MyException(e);
        }
    }


    /**
     * tests the access to simple fiels from a single
     * class, with 2 parameters (mysalary and myname)
     */
    public void testFieldParameter() {
        logger.log(BasicLevel.DEBUG, "SpeedoQL Test - javacc grammar");

        SpeedoQL sql = null;
        logger.log(BasicLevel.DEBUG, "Testing:" + filterFieldParameter);
        sql = new SpeedoQL(new CharArrayReader(filterFieldParameter.toCharArray()));

        try {
            ArrayList all = new ArrayList();
            all.add("org/objectweb/speedo/pobjects/ref/Employee.pd");
            all.add("org/objectweb/speedo/pobjects/ref/Department.pd");
            Manager manager = setupJorm(all);

            SimpleNode node = sql.SpeedoQL();
            SpeedoQLVariableVisitor sqvv
                    = new SpeedoQLVariableVisitor(node, manager);
            sqvv.setLogger(logger);

            toHashtableParams("Float mysalary, String myname", ",");

            sqvv.setParams(hparams);
            sqvv.setVars(null);
            sqvv.setCurrentClass("org.objectweb.speedo.pobjects.ref.Employee");

            sqvv.startVisiting();

            QueryTree queryTree = sqvv.getQueryTree();

            if (logger.isLoggable(BasicLevel.DEBUG)) {
                Field[] fields = queryTree.getTupleStructure().getFields();
                for (int i = 0; i < fields.length; i++) {
                    logger.log(BasicLevel.DEBUG, "i=" + i
                            + ", fieldname=<" + fields[i].getName() + ">");
                }
            }

            SpeedoQLQueryFilterVisitor qfilter =
                    new SpeedoQLQueryFilterVisitor(queryTree, (ASTSpeedoQL) node);
            qfilter.setLogger(logger);

            qfilter.setParams(hparams);
            qfilter.setVars(null);
            qfilter.setCurrentClass("org.objectweb.speedo.pobjects.ref.Employee");
            qfilter.startVisiting();

            // verify the expression is well-formed
            Expression expr = qfilter.getQueryFilter();
            if (expr == null)
                logger.log(BasicLevel.INFO, "ATTENTION: expr null in test");
            else
                logger.log(BasicLevel.INFO, "Expression toString... :" + ExpressionPrinter.e2str(expr));

            SelectProject sp = new SelectProject("", null);
            sp.setQueryFilter(expr);

            queryTree = sp;

            if (logger.isLoggable(BasicLevel.DEBUG)) {
                QueryTreePrinter.printQueryTree(queryTree, logger);
            }
        } catch (ParseException e) {
            throw new MyException(
                    "SpeedoQLMedorTest: Encountered errors during parse.", e);
        } catch (Exception e) {
            throw new MyException(e);
        }
    }

    /**
     * tests the access to a field from a reference. dept field
     * is a field from Employee (Department object reference, which have a name
     * field.
     */

    public void testFieldRef() {
        logger.log(BasicLevel.DEBUG, "SpeedoQL Test - javacc grammar");

        SpeedoQL sql = null;
        logger.log(BasicLevel.DEBUG, "Testing:" + filterFieldRef);
        sql = new SpeedoQL(new CharArrayReader(filterFieldRef.toCharArray()));

        try {
            ArrayList all = new ArrayList();
            all.add("org/objectweb/speedo/pobjects/ref/Employee.pd");
            all.add("org/objectweb/speedo/pobjects/ref/Department.pd");
            Manager manager = setupJorm(all);

            SimpleNode node = sql.SpeedoQL();
            SpeedoQLVariableVisitor sqvv = new SpeedoQLVariableVisitor(node, manager);
            logger.turnOn();
            sqvv.setLogger(logger);

            toHashtableParams("String dep", ",");
            sqvv.setParams(hparams);
            sqvv.setVars(null);
            sqvv.setCurrentClass("org.objectweb.speedo.pobjects.ref.Employee");
            sqvv.startVisiting();

            QueryTree queryTree = sqvv.getQueryTree();

            if (logger.isLoggable(BasicLevel.DEBUG)) {
                Field[] fields = queryTree.getTupleStructure().getFields();
                for (int i = 0; i < fields.length; i++) {
                    logger.log(BasicLevel.DEBUG, "i=" + i
                            + ", fieldname=<" + fields[i].getName() + ">");
                }
            }

            SpeedoQLQueryFilterVisitor qfilter =
                    new SpeedoQLQueryFilterVisitor(queryTree, (ASTSpeedoQL) node);
            qfilter.setLogger(logger);

            qfilter.setParams(hparams);
            qfilter.setVars(null);
            qfilter.setCurrentClass("org.objectweb.speedo.pobjects.ref.Employee");
            qfilter.startVisiting();

            // verify the expression is well-formed
            Expression expr = qfilter.getQueryFilter();
            if (expr == null)
                logger.log(BasicLevel.INFO, "ATTENTION: expr null in test");
            else
                logger.log(BasicLevel.INFO, "Expression toString... :" + ExpressionPrinter.e2str(expr));

            SelectProject sp = new SelectProject("", null);
            sp.setQueryFilter(expr);

            queryTree = sp;

            if (logger.isLoggable(BasicLevel.DEBUG)) {
                QueryTreePrinter.printQueryTree(queryTree, logger);
            }
        } catch (ParseException e) {
            throw new MyException(
                "SpeedoQLMedorTest: Encountered errors during parse.", e);
        } catch (Exception e) {
            throw new MyException(e);
        }
    }

    /**
     * tests the contains method  with a variable: anEmployee, and a
     * parameter: istName
     * Medor: getIn (from NavigatorFactory)
     */
    public void testContainsIn() {
        logger.log(BasicLevel.DEBUG, "SpeedoQL Test - javacc grammar");

        SpeedoQL sql = null;
        logger.log(BasicLevel.DEBUG, "Testing:" + filterIn);
        sql = new SpeedoQL(new CharArrayReader(filterIn.toCharArray()));

        try {
            ArrayList all = new ArrayList();
            all.add("org/objectweb/speedo/pobjects/ref/Employee.pd");
            all.add("org/objectweb/speedo/pobjects/ref/Department.pd");
            Manager manager = setupJorm(all);

            SimpleNode node = sql.SpeedoQL();
            SpeedoQLVariableVisitor sqvv = new SpeedoQLVariableVisitor(node, manager);
            logger.turnOn();
            sqvv.setLogger(logger);

            toHashtableParams("String itsName", ",");

            // variable : "Employee anEmployee" ???
            toHashtableVars("Employee anEmployee", ";");

            sqvv.setParams(hparams);
            sqvv.setVars(hvars);
            sqvv.setCurrentClass("org.objectweb.speedo.pobjects.ref.Department");

            sqvv.startVisiting();

            QueryTree queryTree = sqvv.getQueryTree();

            if (logger.isLoggable(BasicLevel.DEBUG)) {
                Field[] fields = queryTree.getTupleStructure().getFields();
                for (int i = 0; i < fields.length; i++) {
                    logger.log(BasicLevel.DEBUG, "i=" + i
                            + ", fieldname=<" + fields[i].getName() + ">");
                }
            }

            /*
            QueryTree qts[] = sqvv.getQTs();
            // we hope 2
            equals("", qts[0], qts[1]);
            */

            SpeedoQLQueryFilterVisitor qfilter =
                    new SpeedoQLQueryFilterVisitor(queryTree, (ASTSpeedoQL) node);
            qfilter.setLogger(logger);

            qfilter.setParams(hparams);
            qfilter.setVars(hvars);
            qfilter.setCurrentClass("org.objectweb.speedo.pobjects.ref.Department");
            qfilter.startVisiting();

            // verify the expression is well-formed
            Expression expr = qfilter.getQueryFilter();
            if (expr == null)
                logger.log(BasicLevel.INFO, "ATTENTION: expr null in test");
            else
                logger.log(BasicLevel.INFO, "Expression toString... :" + ExpressionPrinter.e2str(expr));


            SelectProject sp = new SelectProject("", null);
            sp.setQueryFilter(expr);

            queryTree = sp;

            QueryTreePrinter.printQueryTree(queryTree, logger);

        } catch (ParseException e) {
            throw new MyException(
                    "SpeedoQLMedorTest: Encountered errors during parse.", e);
        } catch (Exception e) {
            throw new MyException(e);
        }
    }

    /**
     * tests the contains method with a parameter (employee)
     * Medor: getMemberOf (from NavigatorFactory) not yet implemented in MEDOR
     */
    public void testContainsMemberOf() {
        logger.log(BasicLevel.DEBUG, "SpeedoQL Test - javacc grammar");

        SpeedoQL sql = null;
        logger.log(BasicLevel.DEBUG, "Testing:" + filterMemberOf);
        sql = new SpeedoQL(new CharArrayReader(filterMemberOf.toCharArray()));

        try {
            ArrayList all = new ArrayList();
            all.add("org/objectweb/speedo/pobjects/ref/Employee.pd");
            all.add("org/objectweb/speedo/pobjects/ref/Department.pd");
            Manager manager = setupJorm(all);

            SimpleNode node = sql.SpeedoQL();
            SpeedoQLVariableVisitor sqvv = new SpeedoQLVariableVisitor(node, manager);
            logger.turnOn();
            sqvv.setLogger(logger);

            toHashtableParams("Employee employee", ",");

            sqvv.setParams(hparams);
            sqvv.setVars(null);
            sqvv.setCurrentClass("org.objectweb.speedo.pobjects.ref.Employee");

            sqvv.startVisiting();

            QueryTree queryTree = sqvv.getQueryTree();

            if (logger.isLoggable(BasicLevel.DEBUG)) {
                Field[] fields = queryTree.getTupleStructure().getFields();
                for (int i = 0; i < fields.length; i++) {
                    logger.log(BasicLevel.DEBUG, "i=" + i
                            + ", fieldname=<" + fields[i].getName() + ">");
                }
            }

            SpeedoQLQueryFilterVisitor qfilter =
                    new SpeedoQLQueryFilterVisitor(queryTree, (ASTSpeedoQL) node);
            qfilter.setLogger(logger);

            qfilter.setParams(hparams);
            qfilter.setVars(null);
            qfilter.setCurrentClass("org.objectweb.speedo.pobjects.ref.Employee");
            qfilter.startVisiting();

            // verify the expression is well-formed
            Expression expr = qfilter.getQueryFilter();
            if (expr == null)
                logger.log(BasicLevel.INFO, "ATTENTION: expr null in test");
            else
                logger.log(BasicLevel.INFO, "Expression toString... :" + ExpressionPrinter.e2str(expr));


            SelectProject sp = new SelectProject("", null);
            sp.setQueryFilter(expr);

            queryTree = sp;

            QueryTreePrinter.printQueryTree(queryTree, logger);

        } catch (ParseException e) {
            throw new MyException(
                    "SpeedoQLMedorTest: Encountered errors during parse.", e);
        } catch (Exception e) {
            throw new MyException(e);
        }
    }


    ///////////////////////////////////////////////////////////////////////
    // local methods
    ///////////////////////////////////////////////////////////////////////


    private void toHashtableParams(String stringToHash, String separator) {
        StringTokenizer tok = new StringTokenizer(stringToHash, separator);
        hparams = new Hashtable();
        //vparams = new Vector(tok.countTokens());
        while (tok.hasMoreTokens()) {
            String tuple = tok.nextToken().trim();
            int i = tuple.indexOf(' ');
            String name = tuple.substring(i + 1, tuple.length());
            hparams.put(name, new BasicParameterOperand(getPType(tuple.substring(0, i)), name));
        }
    }

    private void toHashtableVars(String stringToHash, String separator) {
        StringTokenizer tok = new StringTokenizer(stringToHash, separator);
        hvars = new Hashtable();

        while (tok.hasMoreTokens()) {
            String tuple = tok.nextToken().trim();
            int i = tuple.indexOf(' ');
            String type = tuple.substring(0, i);
            getPType(type);
            hvars.put(tuple.substring(i + 1, tuple.length()), type);
        }
    }

    /**
     * returns the PType of a given name. If the type does not exist,
     * null is returned.
     * @param name the string representation of the name.
     * @return the corresponding PType of its name.
     */
    private PType getPType(String name) {
        PType type;
        if (name.equals(PTypeSpace.OBJBOOLEAN.getJormName()))
            type = PTypeSpace.OBJBOOLEAN;
        else if (name.equals(PTypeSpace.OBJCHAR.getJormName()))
            type = PTypeSpace.OBJCHAR;
        else if (name.equals(PTypeSpace.OBJBYTE.getJormName()))
            type = PTypeSpace.OBJBYTE;
        else if (name.equals(PTypeSpace.OBJSHORT.getJormName()))
            type = PTypeSpace.OBJSHORT;
        else if (name.equals(PTypeSpace.OBJINT.getJormName()))
            type = PTypeSpace.OBJINT;
        else if (name.equals(PTypeSpace.OBJLONG.getJormName()))
            type = PTypeSpace.OBJLONG;
        else if (name.equals(PTypeSpace.OBJFLOAT.getJormName()))
            type = PTypeSpace.OBJFLOAT;
        else if (name.equals(PTypeSpace.OBJDOUBLE.getJormName()))
            type = PTypeSpace.OBJDOUBLE;
        else if ((name.equals(PTypeSpace.STRING.getJormName())) || (name.compareTo("String") == 0))
            type = PTypeSpace.STRING;
        else if (name.equals(PTypeSpace.DATE.getJormName()))
            type = PTypeSpace.DATE;
        else if (name.equals(PTypeSpace.SERIALIZED.getJormName()))
            type = PTypeSpace.SERIALIZED;
        else // in case we dont know, it is an object
            type = PTypeSpace.REFTOP;
        return type;
    }

    private Manager setupJorm(ArrayList all) {
        JormCompiler jormcompiler = new JormCompiler();
        JormCompilerParameter jcp = jormcompiler.getCompilerParameter();
        JormCompilerConfigurator jcc = jormcompiler.getCompilerConfigurator();
        try {
            jcc.configure();
            jcp.loadConfFile(jcc.getGlobalJormcOptsFile(), jcc.knownMappers());
            jcp.setLoggerFactory(Monolog.loggerFactory);
            jormcompiler.setupLogger();
            jormcompiler.setupMIManager();
            jormcompiler.setupGenerator();
            jcp.projectName = projectName;
            jcp.output = pdPath;
            jcp.bindingInheritance = "%p%c";
            jcp.bindingAbstract = true;
            ArrayList al = new ArrayList(1);
            DTDLocation dtdl = new DTDLocation();
            dtdl.setPublicId("-//ObjectWeb Consortium//DTD JORM DEFINITIONS 1.0//EN");
            dtdl.setLocation(dtdPath);
            al.add(dtdl);
            jcp.dtdLocations = al;
            jcp.inputFiles = all;
            DirJavaExplorer pe = new DirJavaExplorer();
            pe.setLogger(logger);
            pe.addPath(pdPath);
            jcp.classpath = pe;
            Collection c = jcp.getSubMappers("rdb");
            logger.log(BasicLevel.DEBUG, "jormcOpts file:" + jcc.getJormcOptsFile());
            logger.log(BasicLevel.DEBUG, "rdb sub mappers: " + c);

        } catch (PException e) {
            logger.log(BasicLevel.ERROR, "PException in setupJorm (parseFiles)", e);
            fail("Impossible to configure Jorm: " + ExceptionHelper.getNested(e));
        }
        return jormcompiler.getMIManager();
    }

    public class MyException extends RuntimeException {
        Exception inner;

        public MyException(Exception e) {
            super("");
            inner = e;
        }

        public MyException(String s, Exception inner) {
            super(s);
            this.inner = inner;
        }

        public String getMessage() {
            return super.getMessage()
                    + (inner.getMessage() == null ? "" : inner.getMessage());
        }

        public void printStackTrace() {
            inner.printStackTrace();
        }

        public void printStackTrace(PrintStream printStream) {
            inner.printStackTrace(printStream);
        }

        public void printStackTrace(PrintWriter printWriter) {
            inner.printStackTrace(printWriter);
        }
    }
}


